from __future__ import absolute_import
import numpy as np
import math
from .compute_dom import *
from mpi4py import MPI
import os
import os.path
import deepdish as dd


class Material(object):

 def __init__(self,**argv):

  #defaults--
  #argv.setdefault('grid',[50,12,24])

  #MFP discretization-----------------------------
  self.state = {}


  argv.setdefault('model','gray')
  if argv['model'] == 'load':
   #MPI.COMM_WORLD.Barrier()
   if MPI.COMM_WORLD.Get_rank() == 0:
    data = dd.io.load(argv.setdefault('filename','material.hdf5'))
   else: data=None
   self.state.update(MPI.COMM_WORLD.bcast(data,root=0))
   #print(self.state['region'])
  else:

   if argv['model'] == 'nongray':
    self.import_data(argv)

   if argv['model'] == 'gray':
    self.create_gray_model(argv)
   #-----------------------------------------------

   #Angular discretization-------------------------
   if MPI.COMM_WORLD.Get_rank() == 0:
    region = argv.setdefault('region','Matrix')

    dom = compute_dom_3d(argv)
    data = {'dom':dom,'region':region}


   else: data=None
   self.state.update(MPI.COMM_WORLD.bcast(data,root=0))

   #-----------------------------------------------
   #SAVE FILE--------------------
   if argv.setdefault('save',True):
    if MPI.COMM_WORLD.Get_rank() == 0:
     dd.io.save(argv.setdefault('filename','material') + '.hdf5', self.state)


 #def get_region(self):
     #return self.region

 def create_gray_model(self,argv):

  if MPI.COMM_WORLD.Get_rank() == 0:
    data = {}
    mfp = argv.setdefault('mfps',[100e-9])
    n_mfp = len(mfp)
    kappa = argv.setdefault('kappa',1.0)

    data.update({'kappa_bulk_tot':kappa})
    data.update({'mfp_bulk':mfp})
    data.update({'kappa_bulk':[kappa]})
    data.update({'B0':np.array(n_mfp*[1.0/n_mfp])})

    if 'special_point' in argv.keys():
     sp = argv['special_point']
     tmp = np.zeros(n_mfp)
     tmp[sp] = 1.0
     tmp = np.tile(tmp,[n_mfp,1])
     data.update({'B1':tmp})
     data.update({'B2':tmp})
    else:
     data.update({'B1':np.eye(n_mfp)})
     data.update({'B2':np.eye(n_mfp)})


    data.update({'mfp_sampled':mfp})
  else: data=None
  self.state = MPI.COMM_WORLD.bcast(data,root=0)




 def import_data(self,argv):

  if MPI.COMM_WORLD.Get_rank() == 0:

    if '/' in argv['matfile']:
     filename = argv['matfile']
    else:
     filename = os.path.dirname(__file__) + '/materials/'+ argv['matfile']

    data = {}
    #[n_mfp,n_theta,n_phi] = argv['grid']

    n_mfp = int(argv['n_mfp'])
    #READ DATA-------------------------------------------
    tmp = np.loadtxt(filename)
    kappa_bulk_tot = np.sum(tmp[:,1])
    mfp_bulk = tmp[:,0]
    kappa_bulk = tmp[:,1]
    n_mfp_bulk = len(mfp_bulk)
    #-----------------------------------------------------
    delta = argv.setdefault('delta',0.0)
    acc = 0
    mfp_bulk_new = []
    kappa_bulk_new = []
    for m in range(n_mfp_bulk):
     if acc > kappa_bulk_tot*delta and acc < kappa_bulk_tot*(1.0-delta):
      mfp_bulk_new.append(mfp_bulk[m])
      kappa_bulk_new.append(kappa_bulk[m])
     acc += kappa_bulk[m]


    min_mfp = min(mfp_bulk_new)
    min_mfp = min(min_mfp,argv.setdefault('min_mfp',min_mfp))
    max_mfp = max(mfp_bulk_new)
    max_mfp = max(max_mfp,argv.setdefault('max_mfp',max_mfp))

    mfp_sampled_log = np.linspace(np.log10(min_mfp*0.99),np.log10(max_mfp*1.01),n_mfp)

    #---DISCRETIZE MFPS in LOG SPACE---------------------
    mfp_sampled = []
    for m in mfp_sampled_log:
     mfp_sampled.append(pow(10,m))
    #--------------------------------------------------------



    #MFP interpolation---------------------------
    B0 = np.zeros(n_mfp)
    B1 = np.zeros(n_mfp)
    B2 = np.zeros(n_mfp)
    kappa_tot = 0
    for m in range(len(mfp_bulk_new)):
     kappa = kappa_bulk_new[m]
     mfp = mfp_bulk_new[m]
     [m1,m2,f1,f2] = self.get_mfp_interpolation(mfp_sampled_log,np.log10(mfp))
     b0 = kappa
     b1 = kappa/mfp
     b2 = kappa/mfp/mfp
     B0[m1] += f1*b0; B0[m2] += f2*b0
     B1[m1] += f1*b1; B1[m2] += f2*b1
     B2[m1] += f1*b2; B2[m2] += f2*b2
    B0 /=np.sum(B0)
    B1 /=np.sum(B1)
    B2 /=np.sum(B2)
    #---------------------------------------------
    #Compute DOM
    #----------------
    data.update({'kappa_bulk_tot':sum(kappa_bulk_new)})
    data.update({'mfp_bulk':mfp_bulk_new})
    data.update({'kappa_bulk':kappa_bulk_new})
    
    B1 = np.tile(B1,(n_mfp,1))
  
    data.update({'B0':B0})
    data.update({'B1':B1})
    data.update({'B2':np.tile(B2,(n_mfp,1))})
    #data.update({'B1':B1})
    #data.update({'B2':B2})
    data.update({'mfp_sampled':np.array(mfp_sampled)})

  else: data=None
  self.state = MPI.COMM_WORLD.bcast(data,root=0)


 def get_mfp_interpolation(self,mfp_sampled,mfp_bulk) :

   n_mfp = len(mfp_sampled)
   findm = 0
   for m in range(n_mfp-1) :
    if (mfp_bulk >= mfp_sampled[m]) and (mfp_bulk <= mfp_sampled[m+1]) :

     #get data for mfp interpolation
     m1 = m
     m2 = m+1
     mfp1 = mfp_sampled[m1]
     mfp2 = mfp_sampled[m2]
     fm = (mfp_bulk-mfp1)/(mfp2 - mfp1) #for interpolation
     findm = 1
     m_par = [m1,m2,1-fm,fm]
     assert mfp_bulk <= mfp2
     assert mfp_bulk >= mfp1
     assert abs(fm) <= 1
     return m_par

   assert findm == 1



import numpy as np
import sys


def main() :

 data = np.loadtxt('BTE.cumulative_kappa_scalar')
 n_mfp = np.shape(data)[0]

 #Compute dis
 dis = np.zeros(n_mfp)
 dis[0] = data[0,1]
 for m in range(n_mfp-1):
  dis[m+1] = data[m+1,1] - data[m,1]

 #--------------------
 f = open('mat.dat','w+')
 for m in range(n_mfp):
  tmp = data[m,0]*1e-9
  f.write('%10e' %  tmp + '   ')
  f.write('%10e' %  dis[m]+ '\n')
 f.close()

 if __name__ == "__main__":
   main()



import numpy as np
import math
from shapely.geometry import Polygon


def GenerateSquareLatticePores(argv):

  Lx = float(argv['lx'])
  Ly = float(argv['ly'])
  porosity = float(argv['porosity'])

  #Read shape
  shape = argv['shape']
  if shape == 'circle':
     Na = 24; phi0= 0.0

  if shape == 'circle_refined':
     Na = 48; phi0= 0.0

  if shape == 'square':
     Na = 4; phi0 = 0.0
  if shape == 'triangle':
     Na = 3; phi0 = 0.0

  phi0 = argv.setdefault('angle',0.0)
  #Read Number of pores
  Nx = argv.setdefault('Nx',1)
  Ny = argv.setdefault('Ny',1)
  dx = argv.setdefault('dx',0)
  dy = argv.setdefault('dy',0)
  
  frame = []
  frame.append([float(-Lx * Nx)/2,float(Ly * Ny)/2])
  frame.append([float(Lx * Nx)/2,float(Ly * Ny)/2])
  frame.append([float(Lx * Nx)/2,float(-Ly * Ny)/2])
  frame.append([float(-Lx * Nx)/2,float(-Ly * Ny)/2])
  frame_poly = Polygon(frame)

    
  
  #----------------------
  pbc = []
  pbc.append([0,0])
  pbc.append([Lx,0])
  pbc.append([-Lx,0])
  pbc.append([0,Ly])
  pbc.append([0,-Ly])
  pbc.append([Lx,Ly])
  pbc.append([-Lx,-Ly])
  pbc.append([-Lx,Ly])
  pbc.append([Lx,-Ly])

  area = Lx * Ly * porosity

  polygons = []
  for ppp in pbc:
   for kx in range(int(Nx)):
    for ky in range(int(Ny)):
     cx = -Lx*Nx/2.0 + kx*Lx+(0.5+dx)*Lx + ppp[0]
     cy = -Ly*Ny/2.0 + ky*Ly+(0.5+dy)*Ly + ppp[1]
     r = math.sqrt(2.0*area/Na/math.sin(2.0 * np.pi/Na))

     dphi = 2.0*math.pi/Na;
     p = []
     for ka in range(Na):
      ph =  dphi/2 + (ka-1) * dphi + phi0*np.pi/180.0
      px  = cx + r * math.cos(ph)
      py  = cy + r * math.sin(ph)
      p.append([px,py])
     polygons.append(p)

  
  polys_cut = []
  for p in polygons:
   if Polygon(p).intersects(frame_poly):
    polys_cut.append(p)
    

  return frame,polys_cut



import os,sys
import numpy as np
import random
import math
import subprocess


def mesh(argv):


  Lx = argv['lx']
  Ly = argv['ly']
  Lz = argv['lz']



  frame = []
  frame.append([float(-Lx)/2,float(Ly)/2])
  frame.append([float(Lx)/2,float(Ly)/2])
  frame.append([float(Lx)/2,float(-Ly)/2])
  frame.append([float(-Lx)/2,float(-Ly)/2])


  mesh_ext = argv['step']
  Nx = float(argv.setdefault('Nx',1))
  Ny = float(argv.setdefault('Ny',1))
  Nz = float(argv.setdefault('Nz',1))

  Lx *=Nx
  Ly *=Ny
  Lz *=Nz


  store = open('mesh.geo', 'w+')

  points  = [[-Lx/2,-Ly/2,-Lz/2],\
             [-Lx/2,+Ly/2,-Lz/2],\
             [Lx/2,Ly/2,-Lz/2],\
             [Lx/2,-Ly/2,-Lz/2],\
             [-Lx/2,-Ly/2,Lz/2],\
             [-Lx/2,+Ly/2,Lz/2],\
             [Lx/2,Ly/2,Lz/2],\
             [Lx/2,-Ly/2,Lz/2]]

  #Write points
  for p,point in enumerate(points):
   store.write( 'Point('+str(p) +') = {' + str(point[0]) +','+\
                                          str(point[1])+',' +\
                                          str(point[2]) + ','+\
                                          str(mesh_ext) +'};\n')

  #Write lines
  ll = 1
  store.write( 'Line('+str(ll) +') = {' + str(0) +','+ str(1)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(1) +','+ str(2)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(2) +','+ str(3)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(3) +','+ str(0)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(4) +','+ str(5)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(5) +','+ str(6)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(6) +','+ str(7)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(7) +','+ str(4)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(0) +','+ str(4)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(1) +','+ str(5)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(2) +','+ str(6)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(3) +','+ str(7)+'};\n')
  ll += 1

  #Write Surfaces
  loops = [[1,2,3,4],\
           [5,6,7,8],\
           [9,5,-10,-1],\
           [10,6,-11,-2],\
           [12,-7,-11,3],\
           [4,9,-8,-12]]

  for nloop,loop in enumerate(loops) :
   index = nloop + 12
   strc = 'Line Loop(' + str(index)+ ') = {'
   for n,p in enumerate(loop) :
    strc +=str(p)
    if n == len(loop)-1:
     strc += '};\n'
    else :
     strc += ','
   store.write(strc)
   strc = 'Plane Surface(' + str(nloop+1)+ ') = {'+ str(index) + '};\n'
   store.write(strc)

  strc = 'Surface Loop(' + str(12)+ ') = {-1,3,4,5,6,2};\n'
  store.write(strc)
  strc = 'Volume(0)={12};\n'
  store.write(strc)
  strc = 'Physical Volume("Matrix")={0};\n'
  store.write(strc)



  #Write Periodic Surfaces--------------------------------------------
  strc = 'Periodic Surface 1 {1,2,3,4} = 2 {5,6,7,8};\n'
  store.write(strc)
  strc = 'Periodic Surface 3 {9,5,-10,-1} = 5 {12,-7,-11,3};\n'
  store.write(strc)
  strc = 'Periodic Surface 4 {10,6,-11,-2} = 6 {9,-8,-12,4};\n'
  store.write(strc)
  #------------------------------------------------------------------


  boundary_surfaces = []
  if argv.setdefault('Periodic',[True,True,True])[0]:
   strc = r'''Physical Surface('Periodic_1') = {3} ;''' + '\n'
   store.write(strc)
   strc = r'''Physical Surface('Periodic_2') = {5} ;''' + '\n'
   store.write(strc)
  else:
   boundary_surfaces.append(3)
   boundary_surfaces.append(5)

  if argv.setdefault('Periodic',[True,True,True])[1]:
   strc = r'''Physical Surface('Periodic_3') = {4} ;''' + '\n'
   store.write(strc)
   strc = r'''Physical Surface('Periodic_4') = {6} ;''' + '\n'
   store.write(strc)
  else:
   boundary_surfaces.append(4)
   boundary_surfaces.append(6)

  if argv.setdefault('Periodic',[True,True,True])[2]:
   strc = r'''Physical Surface('Periodic_5') = {1} ;''' + '\n'
   store.write(strc)
   strc = r'''Physical Surface('Periodic_6') = {2} ;''' +'\n'
   store.write(strc)
  else:
   boundary_surfaces.append(1)
   boundary_surfaces.append(2)


  #write Boundary surfaces
  if len(boundary_surfaces) > 0:
   strc = r'''Physical Surface('Boundary') = {'''
   for p,side in enumerate(boundary_surfaces) :
    strc +=str(side)
    if p == len(boundary_surfaces)-1:
     strc += '};\n'
    else :
     strc += ','
   store.write(strc)


   store.close()



import numpy as np
import math

def GenerateHexagonalLatticePores(argv):

  a = float(argv.setdefault('a',1))
  porosity = float(argv['porosity'])
  Lx = 2.0 * np.sin(np.pi/6.0)*a
  Ly = 2.0 * np.cos(np.pi/6.0)*a
  phi0 = argv.setdefault('angle',0.0)

  #Read shape
  shape = argv['shape']
  if shape == 'circle':
     Na = 24; #phi0= 360.0/48.0
  if shape == 'square':
     Na = 4; #phi0 = 0.0
  if shape == 'triangle':
     Na = 3; #phi0 = 0.0


  area = 0.5*Lx * Ly * porosity
  r = math.sqrt(2.0*area/Na/math.sin(2.0 * np.pi/Na))

  cc = [[-Lx/2.0,-Ly/2.0],[0.0,0.0],[-Lx/2.0,Ly/2.0],[Lx/2.0,Ly/2.0],[Lx/2.0,-Ly/2.0]]


  polygons = []
  for c in cc:
   dphi = 2.0*math.pi/Na;
   p = []
   for ka in range(Na):
    ph =  dphi/2 + (ka-1) * dphi + phi0*np.pi/180.0
    px  = c[0] + r * math.cos(ph)
    py  = c[1] + r * math.sin(ph)
    p.append([px,py])
   polygons.append(p)

  frame = []
  frame.append([float(-Lx)/2,float(Ly)/2])
  frame.append([float(Lx)/2,float(Ly)/2])
  frame.append([float(Lx)/2,float(-Ly)/2])
  frame.append([float(-Lx)/2,float(-Ly)/2])


  return frame,polygons



from __future__ import print_function
from __future__ import absolute_import

import os
import numpy as np
from scipy.sparse import csc_matrix
from mpi4py import MPI
import shutil
from scipy.sparse.linalg import splu
import matplotlib.pylab as plt
from scipy.sparse import diags
import scipy.io
import deepdish as dd
#from termcolor import colored
from .geometry import Geometry
from .material import Material


class Solver(object):

  def __init__(self,**argv):
   
   self.mesh = Geometry(type='load',filename = argv.setdefault('geometry_filename','geometry'))
   self.ms_error = argv.setdefault('ms_error',0.1)
   self.verbose = argv.setdefault('verbose',True)

   self.n_elems = len(self.mesh.elems)
   self.dim = self.mesh.dim
   self.mfe = False

   self.cache = '.cache'
   self.multiscale = argv.setdefault('multiscale',False)
   self.only_fourier = argv.setdefault('only_fourier',False)
   #self.interface_conductance = argv.setdefault('interface_conductance',1e20)

   #Get material properties---------------------------------
   path = '.'
   files = [i for i in os.listdir(path) if os.path.isfile(os.path.join(path,i)) and \
         'material' in i]

   self.materials = {}
   for filename in files:
    mat = Material(model='load',filename=filename)

    gamma = mat.state['kappa_bulk_tot']/mat.state['mfp_bulk'][0]/mat.state['mfp_bulk'][0]

    mat.state.update({'gamma':gamma})

    self.materials.update({mat.state['region']:mat.state})
    self.B0 = mat.state['B0']
    self.B1 = mat.state['B1']
    self.B2 = mat.state['B2']
    self.kappa_bulk = mat.state['kappa_bulk_tot']
   #-----------------------------------------------------

   
   self.lu = {}

   self.mfp = np.array(mat.state['mfp_sampled'])*1e9 #In nm
   self.n_mfp = len(self.mfp)

   #INITIALIZATION-------------------------
   self.n_el = len(self.mesh.elems)

   self.kappa_factor = self.mesh.kappa_factor

   if self.verbose:
    self.print_logo()

   self.dom = mat.state['dom']
   self.n_theta = self.dom['n_theta']
   self.n_phi = self.dom['n_phi']

   self.mesh.dom = self.dom

   self.print_dof()

   #apply symmetry--------
   if self.dim == 2:
    self.n_theta_irr = int(self.n_theta/2)
    self.symmetry = 2.0
    self.n_index = self.n_phi
   else:
    self.n_theta_irr = 1
    self.symmetry = 1.0
    self.n_index = self.n_phi * self.n_theta
   #---------------------

   #if self.compute_matrix:
   if MPI.COMM_WORLD.Get_rank() == 0:
    directory = self.cache
    if os.path.exists(directory):
     shutil.rmtree(directory,ignore_errors=True)
    os.makedirs(directory)
   MPI.COMM_WORLD.Barrier()
    #compute matrix--------
   argv.setdefault('max_bte_iter',10)

   #if argv['max_bte_iter'] > 0:
       
   self.compute_directional_connections() 
    #tput = compute_sum(self.compute_directional_connections,self.n_index)
   
   

   
   self.compute_space_dependent_kappa_bulk()

   self.assemble_fourier()

   #
   #quit()

   #---------------------------

   #solve the BTE
   self.compute_function(**argv)


   if MPI.COMM_WORLD.Get_rank() == 0:
    if os.path.isdir(self.cache):
     shutil.rmtree(self.cache)

   #SAVE FILE--------------------
   if argv.setdefault('save',True):
    if MPI.COMM_WORLD.Get_rank() == 0:
     dd.io.save(argv.setdefault('filename','solver') + '.hdf5', self.state)

  def compute_space_dependent_kappa_bulk(self):

   if MPI.COMM_WORLD.Get_rank() == 0:

    side_kappa_map = {}
    elem_kappa_map = {}
    region_elem_map = {}
    region_kappa_map = {}
    for ll in self.mesh.side_list['active']:
     (elem1,elem2) = self.mesh.side_elem_map[ll]
     region_1 = self.mesh.get_region_from_elem(elem1)
     kappa_1 = self.materials[region_1]['kappa_bulk_tot']


     if not region_1 in region_kappa_map.keys():
      region_kappa_map.update({region_1:kappa_1})

     if not ll in self.mesh.side_list['Boundary'] :

      region_2 = self.mesh.get_region_from_elem(elem2)
      kappa_2 = self.materials[region_2]['kappa_bulk_tot']

      s =  self.mesh.get_interpolation_weigths(ll,elem1)
      L = self.mesh.get_distance_between_centroids_of_two_elements_from_side(ll)
      L_1 = s*L
      L_2 = (1.0 - s)*L
      if  kappa_1 == 0.0 or kappa_2 == 0.0:
       kappa = 1e-3
      else:
       kappa = L*kappa_1 * kappa_2/(kappa_2 * L_1 + L_2 * kappa_1)# +  kappa_1 * kappa_2/self.interface_conductance)

      #print(kappa_1,kappa_2,kappa)

      side_kappa_map.update({ll:kappa})

      if not elem1 in elem_kappa_map.keys():
       elem_kappa_map.update({elem1:kappa_1})
      if not elem2 in elem_kappa_map.keys():
       elem_kappa_map.update({elem2:kappa_2})

      if not region_2 in region_kappa_map.keys():
       region_kappa_map.update({region_2:kappa_2})

     else:
      (elem1,elem2) = self.mesh.side_elem_map[ll]
      if not elem1 in elem_kappa_map.keys():
        elem_kappa_map.update({elem1:kappa_1})
      side_kappa_map.update({ll:0.0})

    data = {'side_kappa_map':side_kappa_map,'elem_kappa_map':elem_kappa_map,\
              'region_elem_map':region_elem_map,'region_kappa_map':region_kappa_map}

   else: data = None

   data =  MPI.COMM_WORLD.bcast(data,root=0)

   self.side_kappa_map = data['side_kappa_map']
   self.elem_kappa_map = data['elem_kappa_map']
   self.region_elem_map = data['region_elem_map']
   self.region_kappa_map = data['region_kappa_map']




  def compute_directional_diffusion(self,index):

    t = int(index/self.n_phi)
    p = index%self.n_phi
    ss = self.dom['ss'][t][p]
    row_tmp = []
    col_tmp = []
    data_tmp = []
    for kc1,kc2 in zip(*self.mesh.A.nonzero()):
     ll = self.mesh.get_side_between_two_elements(kc1,kc2)
     (v_orth,dummy) = self.mesh.get_decomposed_directions(kc1,kc2,rot=ss)
     row_tmp.append(kc1)
     col_tmp.append(kc2)
     data_tmp.append(self.mesh.get_side_periodic_value(ll,kc2)*v_orth)
    RHS = csc_matrix( (np.array(data_tmp),(np.array(row_tmp),np.array(col_tmp))), shape=(self.n_elems,self.n_elems) )

    scipy.io.mmwrite(self.cache + '/RHS_DIFF_' + str(t) + '_' + str(p) + r'.mtx',RHS)


  def get_material_properties(self,elem):
    region = self.mesh.get_region_from_elem(elem)
    gamma = self.materials[region]['gamma']*1e-18
    mfp = self.materials[region]['mfp_bulk'][0]*1e9 # meter

    return gamma,mfp


  #def compute_directional_connections(self,index,options):
      
  def compute_directional_connections(self):
      
      
   comm = MPI.COMM_WORLD   
   size = comm.size
   rank = comm.rank
   n_index = self.n_index
   block = n_index // size + 1
   for kk in range(block):
    index = rank*block + kk   
    if index < n_index:
  
     r=[];c=[];d=[]
     rk=[]; ck=[]; dk=[]
     P = np.zeros(self.n_elems)
     for i,j in zip(*self.mesh.A.nonzero()):

      #(gamma_i,mfp_i) = self.get_material_properties(i)
      #(gamma_j,mfp_j) = self.get_material_properties(j)

      (cm,cp,dummy,dummy) = self.mesh.get_angular_coeff(i,j,index)
      r.append(i); c.append(i); d.append(cp) #d.append(cp*mfp_i)
      v = self.mesh.get_side_periodic_value(i,j)
      rk.append(i); ck.append(j);
      dk.append(v*(cp+cm)*self.mesh.get_elem_volume(i))

      #dk.append(v*(cp*mfp_i*gamma_i+cm*mfp_j*gamma_j)*self.mesh.get_elem_volume(i))

      r.append(i); c.append(j);  d.append(cm)#d.append(cm*mfp_j*gamma_j/gamma_i)
      
      P[i] += v*cm#*mfp_j*gamma_j/gamma_i

     #------------------------------BOUNDARY-------------------------------
     HW_minus = np.zeros(self.n_elems)
     HW_plus = np.zeros(self.n_elems)
     for elem in self.mesh.boundary_elements:
      (dummy,dummy,cmb,cpb) = self.mesh.get_angular_coeff(elem,elem,index)
      (gamma,mfp) = self.get_material_properties(elem)
      HW_plus[elem] =  cpb/np.pi
      HW_minus[elem] = -cmb#*mfp
     #--------------------------------------------------------------

     #--------------------------WRITE FILES---------------------
     A = csc_matrix( (d,(r,c)), shape=(self.n_elems,self.n_elems) )
     K = csc_matrix( (dk,(rk,ck)), shape=(self.n_elems,self.n_elems) )
     scipy.io.mmwrite(self.cache + '/A_' + str(index) + r'.mtx',A)
     scipy.io.mmwrite(self.cache + '/K_' + str(index) + r'.mtx',K)
     P.dump(open(self.cache + '/P_' + str(index) +r'.np','wb+'))
     HW_minus.dump(open(self.cache +'/HW_MINUS_' + str(index) +r'.np','wb+'))
     HW_plus.dump(open(self.cache + '/HW_PLUS_' + str(index) +r'.np','wb+'))
  MPI.COMM_WORLD.Barrier()


  def solve_bte(self,**argv):

         
   output = self.solve_fourier(**argv) #simple fourier----   
   temperature_fourier = np.tile(output['temperature_fourier'],(self.n_mfp,1))
   temperature_fourier_gradient = np.tile(output['temperature_fourier_gradient'],(self.n_mfp,1,1))
   ms = np.zeros((self.n_mfp,self.n_theta,self.n_phi))
   TB = temperature_fourier.copy()
   TL = temperature_fourier.copy()
   
   TLtot,TBtot  = np.zeros((2,self.n_mfp,self.n_elems))

   kappa_fourier =  np.tile(output['kappa_fourier'],(self.n_mfp))
      
   
   SUP = np.zeros((self.n_mfp,self.n_theta,self.n_phi))
   FLUX = np.zeros((self.n_el,3))
   #ms_1 = 0
   kappa_old = 0.0
   n_iter = 0
   error = 1.0
   comm = MPI.COMM_WORLD  

   block = self.n_index // comm.size + 1
   while n_iter < argv.setdefault('max_bte_iter',10) and \
          error > argv.setdefault('max_bte_error',1e-2):
    
    #print('g')          
    #if comm.rank == 0:
    #   for n in range(self.n_mfp): 
    #    print('ff',min(TL[n]),max(TL[n])) 
    # tmp = self.compute_inho_diffusive_thermal_conductivity(TL[0].copy())
    # print('fff',tmp)
    #if comm.rank == 0:    
    #   fff = TL.copy() 
    #   print(max(fff[0]),max(TL[0]))
      
    
    #Solve fof Fourier-----         
    if n_iter ==  argv.setdefault('max_bte_iter',10)-1:
      #if comm.rank == 0:  
      # for n in range(self.n_mfp):
      #  tmp = self.compute_inho_diffusive_thermal_conductivity(TL[n])  
      #  print(n,tmp)
      
      
      
      #argv.update({'mfe_factor':1.0,\
      #             'verbose':False,\
      #             'lattice_temperature':TL,\
      #             'boundary_temperature':TB.copy(),\
      #             'kappa':np.square(self.mfp),\
      #             'interface_conductance': self.mfp/3.0})    
      #output = self.solve_fourier(**argv) 
      #kappa_fourier = output['kappa_fourier']
      #temperature_fourier = output['temperature_fourier']
      #temperature_fourier_gradient = output['temperature_fourier_gradient']
     #--------------------------------------
      sup_zero = [self.compute_diffusive_thermal_conductivity(TL[n].copy(),mat = np.eye(3)) for n in range(self.n_mfp)]        
      
    comm.Barrier()
    #Output variables---
    SUPp = np.zeros((self.n_mfp,self.n_theta,self.n_phi))
    TLp,TBp  = np.zeros((2,self.n_mfp,self.n_elems))
    FLUXp = np.zeros((self.n_el,3))
    
    rank = comm.rank
    for kk in range(block):
     index = rank*block + kk   
     if index < self.n_index:
      #Read directional information-------------------------------------------------- 
      A = scipy.io.mmread(self.cache + '/A_' + str(index) + '.mtx').tocsc()
      P = np.load(open(self.cache + '/P_' + str(index) +r'.np','rb'))
      HW_MINUS = np.load(open(self.cache + '/HW_MINUS_' + str(index) +r'.np','rb'))
      HW_PLUS = np.load(open(self.cache + '/HW_PLUS_' + str(index) +r'.np','rb'))
      K = scipy.io.mmread(self.cache + '/K_' + str(index) + '.mtx').tocsc()
      #-----------------------------------------------------------------------------

      for tt in range(self.n_theta_irr): #We take into account the symmetry
          
       if self.dim == 2:
        t = tt
        p = index
        theta_factor = self.dom['at'][t] / self.dom['d_theta_vec'][t]
       else:
        t = int(index/self.n_phi)
        p = index%self.n_phi
        theta_factor = 1.0
        
       pre_factor = 3.0/4.0/np.pi*0.5*theta_factor * self.dom['d_omega'][t][p]  
          
       fourier=False
       for m in range(self.n_mfp)[::-1]:
       #----------------------------------------------------------------------------------
        if not fourier:
        
         D = np.multiply(TB[m],HW_MINUS)
         F = scipy.sparse.eye(self.n_elems) + theta_factor  * A * self.mfp[m]
         lu = splu(F.tocsc())
         
         
         RHS = self.mfp[m]*theta_factor * (P + D) + TL[m]

         temp = lu.solve(RHS) 
    
         #gg = self.mesh.compute_grad(temp)
        
         
         #temp =  -np.dot(self.dom['S'][t][p],self.mesh.compute_grad(TL[0]).T)/self.dom['d_omega'][t][p]
         #sup = pre_factor * K.dot(temp).sum()*self.kappa_factor
        
         #sup = self.compute_inho_diffusive_thermal_conductivity(temp,mat=self.dom['ss2'][t][p])*3.0/4.0/np.pi
         #sup = 0.0
         #for i,j in zip(*self.mesh.A.nonzero()):
         #  v = self.mesh.get_side_periodic_value(i,j)            
         #  if v == -1:
         #   ll = self.mesh.get_side_between_two_elements(i,j)
         #   area = self.mesh.get_side_area(ll)   
         #   grad = (gg[i] + gg[j])/2.0
         #   sup -= np.dot(self.dom['ss2'][t][p][0],grad)*self.kappa_factor*3.0/4.0/np.pi*area
         
         sup = pre_factor * K.dot(temp-TL[m]).sum()*self.kappa_factor/self.mfp[m]
         #sup = self.compute_diffusive_thermal_conductivity(temp,mat = np.eye(3))     

         
        
             
         if sup == 0: 
            ms[m,t,p] = 1
         else:
          sup_fourier = kappa_fourier[m]*self.dom['ss2'][t][p][self.mesh.direction][self.mesh.direction]*3.0/4.0/np.pi/self.kappa_bulk
         
          if abs((sup_fourier - sup)/sup) < argv.setdefault('ms_error',1e-2): 
             ms[m,t,p] = 1
             if argv.setdefault('multiscale',False):
              fourier = True
              
        else:
         ms[m,t,p] = 1   
         sup = kappa_fourier[m]*self.dom['ss2'][t][p][self.mesh.direction][self.mesh.direction]*3.0/4.0/np.pi/self.kappa_bulk
         temp = temperature_fourier[m]-self.mfp[m]*np.dot(self.dom['S'][t][p],temperature_fourier_gradient[m])/self.dom['d_omega'][t][p]
         
        SUPp[m,t,p] += sup
        
        TLp += np.outer(self.B2[:,m],temp)*self.dom['d_omega'][t][p]*self.symmetry/4.0/np.pi    
        
        #for n in range(self.n_mfp):
        # TLp[n] += temp*self.dom['d_omega'][t][p]*self.symmetry/4.0/np.pi * self.B2[m,0]   
        
        
       #--------------------------
        if self.symmetry == 2.0:
         FLUXp += np.mean(self.B1[:,m])*np.outer(temp,np.multiply(self.dom['S'][t][p],[2.0,2.0,0.0]))
         SUPp[m,self.n_theta -t -1,p] += SUPp[m,t,p]
         
         #TBp += self.dom['at'][t]*self.symmetry*np.outer(self.B1[:,m],np.multiply(temp,HW_PLUS))
         TBp += self.dom['at'][t]*self.symmetry*np.outer(self.B1[:,m],np.multiply(temp,HW_PLUS))
        else:
         FLUXp += np.mean(self.B1[:,m])*np.outer(temp,self.dom['S'][t][p])
         TBp[m] += np.multiply(temp,HW_PLUS)*self.symmetry


   
    comm.Barrier()
    comm.Allreduce([SUPp,MPI.DOUBLE],[SUP,MPI.DOUBLE],op=MPI.SUM)
    comm.Allreduce([TLp,MPI.DOUBLE],[TL,MPI.DOUBLE],op=MPI.SUM)
    comm.Allreduce([TBp,MPI.DOUBLE],[TB,MPI.DOUBLE],op=MPI.SUM)
    comm.Allreduce([FLUXp,MPI.DOUBLE],[FLUX,MPI.DOUBLE],op=MPI.SUM)
    #TL = TLtot.copy()
    #TB = TBtot.copy()
    #TL = TLtot.copy()
    #TB = TLtot.copy()    
    #  print('gg')  
    #  for n in range(self.n_mfp):  
          
    #   print(max(TL[n]))  
    #print('gg')
        
    kappa = sum([self.B0[m]*SUP[m,:,:].sum() for m in range(np.shape(SUP)[0])])*self.kappa_bulk
    sup_m = [SUP[m,:,:].sum() for m in range(np.shape(SUP)[0])]
    n_iter +=1
    error = abs(kappa-kappa_old)/abs(kappa)
    kappa_old = kappa
    if rank==0:
        
     print(kappa)
    #print(rank,np.linalg.norm(TL_old[0]-TL[0]))
    
    #check---
    
    
    
    #if rank == 0:
    # [print(max(TLtot[u])) for u in range(self.n_mfp)]
    #quit() 
    #Solve fourier---------------------------------------------------
    #print(max(TL[0]))
    #if n_iter < argv.setdefault('max_bte_iter',10) and \
    #      error > argv.setdefault('max_bte_error',1e-2) and \
    #      argv.setdefault('multiscale',False):
          
    # argv.update({'mfe_factor':1.0,\
    #              'verbose':False,\
    #              'lattice_temperature':TL,\
    #              'boundary_temperature':TB,\
    #              'kappa':1.0/3.0*np.square(self.mfp),\
    #              'interface_conductance': 1.0/2.0/self.mfp})    
    # output = self.solve_fourier(**argv) 
    # kappa_fourier = output['kappa_fourier']
    # temperature_fourier = output['temperature_fourier']
    # temperature_fourier_gradient = output['temperature_fourier_gradient']
    #----------------------------------------------------------------
    
    #if rank==0:
    #  print(kappa)  
    #  print(sum(sum(sum(ms)))/float(self.n_mfp*self.n_theta*self.n_phi))
      #print(error)
     #print(ms_1) 
     #print('{0:7d} {1:20.4E} {2:25.4E} {4:8.1E} {5:8.1E}'.format(n_iter,kappa,error,ms_1,1.0-ms_1))
  
   #argv.update({'mfe_factor':1.0,\
   #               'verbose':False,\
   #               'lattice_temperature':TL_old,\
   #               'boundary_temperature':TB_old,\
   #               'kappa':np.square(self.mfp)/3.0,\
   #               'interface_conductance': self.mfp/4.0})    
   #output = self.solve_fourier(**argv) 
   #kappa_fourier = output['kappa_fourier']
   #if rank==0:  
    #print(max(TL_old[0])) 
    #print(max(TL_old[-1])) 

    
   # for n in range(self.n_mfp):
   #  tmp = self.compute_inho_diffusive_thermal_conductivity(TL_old[n])
    # print(tmp)
  
    #print(tmp)
   if rank == 0:
     plt.plot(self.mfp,sup_m,'g')  
     plt.plot(self.mfp,sup_zero,'k')  

     #plt.plot(self.mfp,kappa_fourier/self.kappa_bulk,'r')
     plt.xscale('log')
     plt.ylim([0,1])
     plt.show()
     
    
    #sup_m = [SUP[m,:,:].sum() for m in range(np.shape(SUP)[0])]
    #output = {'suppression': SUP,\
    #        'suppression_mfp':sup_m,\
    #        'ThermalFlux_BTE': FLUX,\
    #        'Temperature_BTE': TL[0],\
    #        'kappa_bte':kappa}
  
    #return output
    

  def print_bulk_kappa(self):
      
    if MPI.COMM_WORLD.Get_rank() == 0:
     if self.verbose:
      print('  ')
      for region in self.region_kappa_map.keys():
       print(region.ljust(15) +  '{:8.2f}'.format(self.region_kappa_map[region])+ ' W/m/K')    

  def compute_function(self,**argv):      
    
      
   #Bulk
   self.print_bulk_kappa()
   
   #First guess: fourier
   
   
   #BTE
   self.solve_bte(**argv)

   quit()
  

  def assemble_fourier(self) :

   if  MPI.COMM_WORLD.Get_rank() == 0:
    row_tmp = []
    col_tmp = []
    data_tmp = []
    row_tmp_b = []
    col_tmp_b = []
    data_tmp_b = []
    data_tmp_b2 = []
    B = np.zeros(self.n_elems)
    for kc1,kc2 in zip(*self.mesh.A.nonzero()):
     #ll = self.mesh.get_side_between_two_elements(kc1,kc2)

     #kappa = self.side_kappa_map[ll]
     kappa = 1.0
     #kappa = 100.0
     #print(kappa)        print(KAPPAp)


     (v_orth,dummy) = self.mesh.get_decomposed_directions(kc1,kc2)
     vol1 = self.mesh.get_elem_volume(kc1)
     #vol2 = self.mesh.get_elem_volume(kc2)
     row_tmp.append(kc1)
     col_tmp.append(kc2)
     data_tmp.append(-v_orth/vol1*kappa)
     row_tmp.append(kc1)
     col_tmp.append(kc1)
     data_tmp.append(v_orth/vol1*kappa)
     #----------------------
     row_tmp_b.append(kc1)
     col_tmp_b.append(kc2)
     data_tmp_b.append(self.mesh.get_side_periodic_value(kc2,kc1)*v_orth*kappa)
     data_tmp_b2.append(self.mesh.get_side_periodic_value(kc2,kc1))
     #---------------------
     B[kc1] += self.mesh.get_side_periodic_value(kc2,kc1)*v_orth/vol1*kappa

    #Boundary_elements
    FF = np.zeros(self.n_elems)
    boundary_mask = np.zeros(self.n_elems)
    for side in self.mesh.side_list['Boundary']:
     elem = self.mesh.side_elem_map[side][0]
     vol = self.mesh.get_elem_volume(elem)
     area = self.mesh.get_side_area(side)
     FF[elem] = area/vol
     boundary_mask[elem] = 1.0

    F = csc_matrix((np.array(data_tmp),(np.array(row_tmp),np.array(col_tmp))), shape=(self.n_elems,self.n_elems) )
    RHS = csc_matrix( (np.array(data_tmp_b),(np.array(row_tmp_b),np.array(col_tmp_b))), shape=(self.n_elems,self.n_elems) )
    PER = csc_matrix( (np.array(data_tmp_b2),(np.array(row_tmp_b),np.array(col_tmp_b))), shape=(self.n_elems,self.n_elems) )

    #-------
    data = {'F':F,'RHS':RHS,'B':B,'PER':PER,'boundary_mask':boundary_mask,'FF':FF}
   else: data = None
   data =  MPI.COMM_WORLD.bcast(data,root=0)
   self.F = data['F']
   self.RHS = data['RHS']
   self.PER = data['PER']
   self.B = data['B']
   self.FF = data['FF']
   self.boundary_mask = data['boundary_mask']


  def solve_fourier(self,**argv):
       
        
    #Update matrices---------------------------------------------------
    G = argv.setdefault('interface_conductance',np.zeros(1))*self.dom['correction']
    TL = argv.setdefault('lattice_temperature',np.zeros((1,self.n_el)))
    TB = argv.setdefault('boundary_temperature',np.zeros((1,self.n_el)))
    mfe_factor = argv.setdefault('mfe_factor',0.0)
    kappa_bulk = argv.setdefault('kappa',[1.0])
   
    #print(kappa_bulk)
    #G *=5.0
    n_index = len(kappa_bulk)
    G = np.zeros(n_index)
    #-------------------------------
    KAPPAp,KAPPA = np.zeros((2,n_index))
    FLUX,FLUXp = np.zeros((2,n_index,3,self.n_elems))
    TLp,TLtot = np.zeros((2,n_index,self.n_elems))

    
    #set up MPI------------------
    comm = MPI.COMM_WORLD  
    size = comm.size
    rank = comm.rank
    
    
    block = n_index // size + 1
    for kk in range(block):
     index = rank*block + kk   
     if index < n_index:
      #Aseemble matrix-------------       
     
      #diag_1 = diags([self.FF],[0]).tocsc()
      #A = (self.F + diag_1*G[index])   + mfe_factor * csc_matrix(scipy.sparse.eye(self.n_elems))/kappa_bulk[index]
      #B = (self.B + np.multiply(self.FF,TB[index])*G[index])  + mfe_factor * TL[index]/kappa_bulk[index] 
      A  = kappa_bulk[index] * self.F + mfe_factor * csc_matrix(scipy.sparse.eye(self.n_elems)) 
      A += mfe_factor * diags([self.FF],[0]).tocsc() * G[index]  
      
      B  = kappa_bulk[index] * self.B + mfe_factor * TL[index] 
      B += mfe_factor * np.multiply(self.FF,TB[index]) * G[index]
      
      if mfe_factor == 0.0: A[10,10] = 1.0
      SU = splu(A)
      C = np.zeros(self.n_el)
      #--------------------------------------

      n_iter = 0
      kappa_old = 0
      error = 1        

      while error > argv.setdefault('max_fourier_error',1e-11) and \
                    n_iter < argv.setdefault('max_fourier_iter',10) :

        RHS = B + C*kappa_bulk[index]
        if mfe_factor == 0.0: RHS[10] = 0.0      
        temp = SU.solve(RHS)
        temp = temp - (max(temp)+min(temp))/2.0
        (C,flux) = self.compute_non_orth_contribution(temp)
        KAPPAp[index] = self.compute_inho_diffusive_thermal_conductivity(temp)
        error = abs((KAPPAp[index] - kappa_old)/KAPPAp[index])
        kappa_old = KAPPAp[index]
        n_iter +=1
        
      #FLUXp = np.array([-self.elem_kappa_map[k]*tmp for k,tmp in enumerate(flux)])
      FLUXp[index] = flux.T
      TLp[index] = temp
      
    comm.Allreduce([KAPPAp,MPI.DOUBLE],[KAPPA,MPI.DOUBLE],op=MPI.SUM)
    comm.Allreduce([TLp,MPI.DOUBLE],[TLtot,MPI.DOUBLE],op=MPI.SUM) #to be changed
    comm.Allreduce([FLUXp,MPI.DOUBLE],[FLUX,MPI.DOUBLE],op=MPI.SUM) #
 
    if rank== 0 and argv.setdefault('verbose',True):
     print('  ')
     print('Thermal Conductivity: '.ljust(20) +  '{:8.2f}'.format(KAPPA[0])+ ' W/m/K')
     print('  ')
    


    return {'kappa_fourier':KAPPA,'temperature_fourier_gradient':FLUX,'temperature_fourier':TLtot}

      
  def compute_non_orth_contribution(self,temp) :

    self.gradT = self.mesh.compute_grad(temp)
    C = np.zeros(self.n_el)
    for i,j in zip(*self.mesh.A.nonzero()):

     if not i==j:
      #ll = self.mesh.get_side_between_two_elements(i,j)
      #kappa = self.side_kappa_map[ll]
      #Get agerage gradient----
      side = self.mesh.get_side_between_two_elements(i,j)


      w = self.mesh.get_interpolation_weigths(side,i)
      grad_ave = w*self.gradT[i] + (1.0-w)*self.gradT[j]
      #------------------------
      (dumm,v_non_orth) = self.mesh.get_decomposed_directions(i,j)

      
      C[i] += np.dot(grad_ave,v_non_orth)/2.0/self.mesh.get_elem_volume(i)#*kappa
      C[j] -= np.dot(grad_ave,v_non_orth)/2.0/self.mesh.get_elem_volume(j)#*kappa


    return C, self.gradT# [-self.elem_kappa_map[n]*tmp for n,tmp in enumerate(self.gradT)]


  def compute_inho_diffusive_thermal_conductivity(self,temp,mat=np.eye(3)):

   kappa = 0
   for i,j in zip(*self.PER.nonzero()):

    ll = self.mesh.get_side_between_two_elements(i,j)
    kappa_b = self.side_kappa_map[ll]
    (v_orth,dummy) = self.mesh.get_decomposed_directions(i,j,rot=mat)
    
    kappa += 0.5*v_orth *self.PER[i,j]*(temp[j]+self.PER[i,j]-temp[i])*kappa_b

   return kappa*self.kappa_factor


  def compute_diffusive_thermal_conductivity(self,temp,mat=np.eye(3)):

   kappa = 0
   for i,j in zip(*self.PER.nonzero()):
    (v_orth,dummy) = self.mesh.get_decomposed_directions(i,j,rot=mat)
    kappa += 0.5*v_orth *self.PER[i,j]*(temp[j]+self.PER[i,j]-temp[i])

   return kappa*self.kappa_factor

  def print_logo(self):

   if MPI.COMM_WORLD.Get_rank() == 0:
    print(' ')
    print(r'''  ___                   ____ _____ _____ ''' )
    print(r''' / _ \ _ __   ___ _ __ | __ )_   _| ____|''')
    print(r'''| | | | '_ \ / _ \ '_ \|  _ \ | | |  _|  ''')
    print(r'''| |_| | |_) |  __/ | | | |_) || | | |___ ''')
    print(r''' \___/| .__/ \___|_| |_|____/ |_| |_____|''')
    print(r'''      |_|                                ''')
    print('')
    print('Giuseppe Romano [romanog@mit.edu]')
    print(' ')

  def print_dof(self):

   if MPI.COMM_WORLD.Get_rank() == 0:
    if self.verbose:
     print(' ')
     print('Elements:          ' + str(self.n_el))
     print('Azimuthal angles:  ' + str(self.n_theta))
     print('Polar angles:      ' + str(self.n_phi))
     print('Mean-free-paths:   ' + str(self.n_mfp))
     print('Degree-of-freedom: ' + str(self.n_mfp*self.n_theta*self.n_phi*self.n_el))



      #Thermalize boundaries--------------------------
      #Hot = np.zeros(self.n_elems)
      #for side in self.mesh.side_list['Hot']:
   #    (coeff,elem) = self.mesh.get_side_coeff(side)#
   #    tmp = np.dot(coeff,angle_factor)#
   #    if tmp < 0:
   #     Hot[elem] = -tmp*0.5#
   #    else:
   #     r.append(elem); c.append(elem); d.append(tmp)

    #  Cold = np.zeros(self.n_elems)
    #  for side in self.mesh.side_list['Cold']:#
       #(coeff,elem) = self.mesh.get_side_coeff(side)
       #tmp = np.dot(coeff,angle_factor)
       #if tmp < 0:
       # Cold[elem] = tmp*0.5
       #else:
       # r.append(elem); c.append(elem); d.append(tmp)

      #-----------------------------------------------
  #Hot.dump(open(self.cache + '/Hot_' + str(index) +r'.np','w+'))
  #Cold.dump(open(self.cache +'/Cold_' + str(index) +r'.np','w+'))



from __future__ import absolute_import

import os
import numpy as np
#import time
from scipy.sparse import csc_matrix
from mpi4py import MPI
import shutil
from scipy.sparse.linalg import splu
#from scipy.sparse.linalg import lgmres
import matplotlib.pylab as plt
from scipy.sparse import diags
import scipy.io
import deepdish as dd
#from termcolor import colored
from .geometry import Geometry
from .material import Material
from scipy import interpolate

def log_interp1d(xx, yy, kind='linear'):
    
     scale = min(yy)
     yy -= min(yy)
     yy+= 1e-12
     logx = np.log10(xx)
     logy = np.log10(yy)
     lin_interp = interpolate.interp1d(logx,logy,kind=kind)
     log_interp = lambda zz: np.power(10.0, lin_interp(np.log10(zz)))  +scale -1e-12
     
     return log_interp


class Solver(object):

  def __init__(self,**argv):
   
   self.mesh = Geometry(type='load',filename = argv.setdefault('geometry_filename','geometry'))
   self.ms_error = argv.setdefault('ms_error',0.1)
   self.verbose = argv.setdefault('verbose',True)

   self.n_elems = len(self.mesh.elems)
   self.dim = self.mesh.dim
   self.mfe = False

   self.cache = '.cache'
   self.multiscale = argv.setdefault('multiscale',False)
   self.only_fourier = argv.setdefault('only_fourier',False)
   #self.interface_conductance = argv.setdefault('interface_conductance',1e20)

   #Get material properties---------------------------------
   #path = '.'
   #files = [i for i in os.listdir(path) if os.path.isfile(os.path.join(path,i)) and \
   #      'material' in i]

   #self.materials = {}
   #for filename in files:
   mat = Material(model='load',filename='material.hdf5')
   #print(mat.state)
   #data = dd.io.load(argv.setdefault('filename','material.hdf5'))

  
   self.J0 = mat.state['J0']
   self.J1 = mat.state['J1']
   self.J2 = mat.state['J2']
   self.J3 = mat.state['J3']
   self.kappa_bulk = mat.state['kappa_bulk']
   
   self.kappa_bulk_tot = mat.state['kappa_bulk_tot']
   self.mfp_bulk = mat.state['mfp_bulk']*1e9
   self.rmfp = mat.state['rmfp']
   

   self.trials = mat.state['trials']
   self.n_mfp_bulk = len(self.mfp_bulk)
   #self.dom = mat.state['dom']   
    #gamma = mat.state['kappa_bulk_tot']/mat.state['mfp_bulk'][0]/mat.state['mfp_bulk'][0]
    #mat.state.update({'gamma':gamma})
    #self.materials.update({mat.state['region']:mat.state})
    #self.B0 = mat.state['B0']
    #self.B1 = mat.state['B1']
    #self.J = mat.state['J']
    #self.B2 = mat.state['B2']
    #self.factor = mat.state['factor']
    #self.kappa_bulk = mat.state['kappa_bulk']
    #self.kappa_bulk_tot = mat.state['kappa_bulk_tot']
    #self.mfp_bulk = mat.state['mfp_bulk']*1e9
   #-----------------------------------------------------
   
   

   
   
   self.multiscale = argv.setdefault('multiscale',False)

   self.rmfp = np.array(mat.state['rmfp']) #In nm
   self.n_rmfp = len(self.rmfp)

   #INITIALIZATION-------------------------
   self.n_el = len(self.mesh.elems)

   self.kappa_factor = self.mesh.kappa_factor

   if self.verbose:
    self.print_logo()

   self.dom = mat.state['dom']
   self.n_theta = self.dom['n_theta']
   self.n_phi = self.dom['n_phi']

   self.mesh.dom = self.dom

   self.print_dof()

   # self.n_index = self.n_phi * self.n_theta
   #---------------------

   #if self.compute_matrix:
   #directory = self.cache
   #if os.path.exists(directory):
   #  shutil.rmtree(directory,ignore_errors=True)
   #os.makedirs(directory)
   argv.setdefault('max_bte_iter',10)

   #if argv['max_bte_iter'] > 0:
   
   
   if self.dim == 2 :
    self.n_index = self.n_phi
    self.compute_directional_connections_2D()
   if self.dim == 3 :
    self.n_index = self.n_phi*self.n_theta
    self.compute_directional_connections_3D()   
    
   #self.compute_space_dependent_kappa_bulk()
   self.assemble_fourier()

   #---------------------------
   
   #solve the BTE
   
   self.compute_function(**argv)
   quit()


   if MPI.COMM_WORLD.Get_rank() == 0:
    if os.path.isdir(self.cache):
     shutil.rmtree(self.cache)

   #SAVE FILE--------------------
   if argv.setdefault('save',True):
    if MPI.COMM_WORLD.Get_rank() == 0:
     dd.io.save(argv.setdefault('filename','solver') + '.hdf5', self.state)


  def compute_directional_connections_3D(self):

    for index in range(self.n_index):

     t = int(index/self.n_phi)
     p = index%self.n_phi   
        
     angle = self.dom['polar_dir'][t]*self.dom['fphi']*self.dom['ftheta']   

     
     r=[];c=[];d=[]
     rk=[]; ck=[]; dk=[]
     P = np.zeros(self.n_elems)
     HW_minus = np.zeros(self.n_elems)
     HW_plus = np.zeros(self.n_elems)
     for i,j in zip(*self.mesh.A.nonzero()):
      (cm,cp,cmb,cpb) = self.mesh.get_angular_coeff_old(i,j,angle)
      r.append(i); c.append(i); d.append(cp) 
      v = self.mesh.get_side_periodic_value(i,j)
      rk.append(i); ck.append(j);
      dk.append(v*(cp+cm)*self.mesh.get_elem_volume(i)*0.5)
      r.append(i); c.append(j);  d.append(cm)
      P[i] += v*cm
      HW_plus[i]  += cpb
      HW_minus[i] -= cmb
     #--------------------------------------------------------------
     #--------------------------WRITE FILES---------------------
     A = csc_matrix( (d,(r,c)), shape=(self.n_elems,self.n_elems) )
     K = csc_matrix( (dk,(rk,ck)), shape=(self.n_elems,self.n_elems) )
     scipy.io.mmwrite(self.cache + '/A_' + str(index) + r'.mtx',A)
     scipy.io.mmwrite(self.cache + '/K_' + str(index) + r'.mtx',K)
     P.dump(open(self.cache + '/P_' + str(index) +r'.np','wb+'))
     HW_minus.dump(open(self.cache +'/HW_MINUS_' + str(index) +r'.np','wb+'))
     HW_plus.dump(open(self.cache + '/HW_PLUS_' + str(index) +r'.np','wb+'))


  def compute_directional_connections_2D(self):

    for index in range(self.n_index):

        
     angle = self.dom['polar_dir'][index]*self.dom['fphi']   
     r=[];c=[];d=[]
     rk=[]; ck=[]; dk=[]
     P = np.zeros(self.n_elems)
     HW_minus = np.zeros(self.n_elems)
     HW_plus = np.zeros(self.n_elems)
     
     for i,j in zip(*self.mesh.A.nonzero()):
      (cm,cp,cmb,cpb) = self.mesh.get_angular_coeff_old(i,j,angle)
      r.append(i); c.append(i); d.append(cp) 
      v = self.mesh.get_side_periodic_value(i,j)
      rk.append(i); ck.append(j);
      dk.append(v*(cp+cm)*self.mesh.get_elem_volume(i)*0.5)
      r.append(i); c.append(j);  d.append(cm)
      P[i] += v*cm
      HW_plus[i]  += cpb
      HW_minus[i] -= cmb
     #--------------------------------------------------------------
     #--------------------------WRITE FILES---------------------
     A = csc_matrix( (d,(r,c)), shape=(self.n_elems,self.n_elems) )
     K = csc_matrix( (dk,(rk,ck)), shape=(self.n_elems,self.n_elems) )
     scipy.io.mmwrite(self.cache + '/A_' + str(index) + r'.mtx',A)
     scipy.io.mmwrite(self.cache + '/K_' + str(index) + r'.mtx',K)
     P.dump(open(self.cache + '/P_' + str(index) +r'.np','wb+'))
     HW_minus.dump(open(self.cache +'/HW_MINUS_' + str(index) +r'.np','wb+'))
     HW_plus.dump(open(self.cache + '/HW_PLUS_' + str(index) +r'.np','wb+'))


      
  def solve_bte(self,**argv):


   #Variable declaration-----   
   
   argv.update({'kappa':[self.kappa_bulk_tot]})
   output = self.solve_fourier(**argv) #simple fourier----   
  
   temperature_fourier = np.tile(output['temperature_fourier'],(self.n_rmfp,1))
   #temperature_fourier_gradient = np.tile(output['temperature_fourier_gradient'],(self.n_rmfp,1,1))
   #ms = np.zeros((self.n_mfp,self.n_theta,self.n_phi))
   TB = temperature_fourier.copy()
   TL = temperature_fourier.copy()
         
   #kappa_fourier =  np.tile(output['kappa_fourier'],(self.n_rmfp))
   
    
   #FLUX = np.zeros((self.n_el,3))
   kappa_old = 0.0
   n_iter = 0
   error = 1.0
   kappa_mfe = np.square(self.rmfp)/2.0
   
   comm = MPI.COMM_WORLD  
   block = self.n_index // comm.size + 1
   lu = {}

   rank = comm.rank 


  
   while n_iter < argv.setdefault('max_bte_iter',10) and \
          error > argv.setdefault('max_bte_error',1e-2):

              
    #solve_fourier
    #if n_iter ==  argv.setdefault('max_bte_iter',10)-1:
     

    Jp,J = np.zeros((2,self.n_rmfp,self.n_elems))
    kernelp,kernel = np.zeros((2,self.n_rmfp))
    Tp,T = np.zeros((2,self.n_rmfp,self.n_elems))
    
    for kk in range(block):
     index = rank*block + kk   
     if index < self.n_index:
        
      #Read directional information-------------------------------------------------- 
      A = scipy.io.mmread(self.cache + '/A_' + str(index) + '.mtx').tocsc()
      P = np.load(open(self.cache + '/P_' + str(index) +r'.np','rb'))
      HW_MINUS = np.load(open(self.cache + '/HW_MINUS_' + str(index) +r'.np','rb'))
      HW_PLUS = np.load(open(self.cache + '/HW_PLUS_' + str(index) +r'.np','rb'))
      K = scipy.io.mmread(self.cache + '/K_' + str(index) + '.mtx').tocsc()
       
      for n in range(self.n_rmfp)[::-1]:
          
       local_index = index*self.n_rmfp + n   
       #Solve the system--------    
       RHS = self.rmfp[n]  * (P + np.multiply(TB[n],HW_MINUS)) + TL[n]      
       if not local_index in lu.keys():   
        F = scipy.sparse.eye(self.n_elems) +  A * self.rmfp[n] 
        lu.update({local_index:splu(F.tocsc())})   
       temp = lu[local_index].solve(RHS)        
       #--------------------------------
      
       Tp[n] += temp
       kernelp[n] += 3*2/4/np.pi*self.kappa_factor*K.dot(temp-TL[n]).sum()*self.dom['fphi']*self.dom['dphi'] 
       Jp[n] += 2.0*np.multiply(temp,HW_PLUS)
    
    
    comm.Allreduce([Tp,MPI.DOUBLE],[T,MPI.DOUBLE],op=MPI.SUM)
    comm.Allreduce([kernelp,MPI.DOUBLE],[kernel,MPI.DOUBLE],op=MPI.SUM)
    comm.Allreduce([Jp,MPI.DOUBLE],[J,MPI.DOUBLE],op=MPI.SUM)
    
    n_iter +=1
    
    
    #Solve Fourier-----------------------------------------------
    if self.multiscale:
     argv.update({'mfe_factor':1.0,\
                  'verbose':False,\
                  'lattice_temperature':TL.copy(),\
                  'boundary_temperature':TB.copy(),\
                  'kappa':kappa_mfe,\
                  'interface_conductance': self.rmfp/2})    
     output = self.solve_fourier(**argv) 
     kappa_fourier = output['kappa_fourier']
    #----------------------------------------------------------- 
    
    
    #Boundary temperature
    
    
  
    
    TB = np.tile([np.sum(np.multiply(self.J1,log_interp1d(self.rmfp,J.T[e])(self.trials))) \
         for e in range(self.n_elems)],(self.n_rmfp,1))*self.dom['dphi']/np.pi
    
    
    
    #Lattice temperature   
    TL = self.dom['dphi']/2.0/np.pi*np.tile([np.sum(np.multiply(self.J2,log_interp1d(self.rmfp,T.T[e])(self.trials))) \
        for e in range(self.n_elems)],(self.n_rmfp,1))
   
    #Suppression function   
    SUP = np.sum(np.multiply(self.J3,log_interp1d(self.rmfp,kernel.clip(min=1e-12))(self.trials)),axis=1)    
    
    kappa = np.dot(self.kappa_bulk,SUP)
   

    error = abs(kappa-kappa_old)/abs(kappa)
    
    kappa_old = kappa
     
    
       
    #Thermal conductivity   
    if rank==0:
     print(kappa)       
     
   if rank==0:
    
    if self.multiscale:
     SUP_DIF = np.sum(np.multiply(self.J0,log_interp1d(self.rmfp,kappa_fourier/kappa_mfe)(self.trials)),axis=1)   
     plt.plot(self.mfp_bulk,SUP_DIF,'r') 
    plt.plot(self.mfp_bulk,SUP) 
    #
    plt.xscale('log')
    plt.show()
    
   
    
    
    
    
    

  def print_bulk_kappa(self):
      
    if MPI.COMM_WORLD.Get_rank() == 0:
     if self.verbose:
      #print('  ')
      #for region in self.region_kappa_map.keys():
       print('{:8.2f}'.format(self.kappa_bulk_tot)+ ' W/m/K')    
       #print(region.ljust(15) +  '{:8.2f}'.format(self.region_kappa_map[region])+ ' W/m/K')    

  def compute_function(self,**argv):      
    
      
   #Bulk
   self.print_bulk_kappa()
   
   #First guess: fourier
   
   
   #BTE
   self.solve_bte(**argv)

   quit()
  

  def assemble_fourier(self) :

   if  MPI.COMM_WORLD.Get_rank() == 0:
    row_tmp = []
    col_tmp = []
    data_tmp = []
    row_tmp_b = []
    col_tmp_b = []
    data_tmp_b = []
    data_tmp_b2 = []
    B = np.zeros(self.n_elems)
    kappa_mask = []
    for kc1,kc2 in zip(*self.mesh.A.nonzero()):
     #ll = self.mesh.get_side_between_two_elements(kc1,kc2)

     #kappa = self.side_kappa_map[ll]
     kappa = 1.0
     #kappa = 100.0
     #print(kappa)        print(KAPPAp)


     (v_orth,dummy) = self.mesh.get_decomposed_directions(kc1,kc2)
     vol1 = self.mesh.get_elem_volume(kc1)
     #vol2 = self.mesh.get_elem_volume(kc2)
     row_tmp.append(kc1)
     col_tmp.append(kc2)
     data_tmp.append(-v_orth/vol1*kappa)
     row_tmp.append(kc1)
     col_tmp.append(kc1)
     data_tmp.append(v_orth/vol1*kappa)
     #----------------------
     row_tmp_b.append(kc1)
     col_tmp_b.append(kc2)
     tt = self.mesh.get_side_periodic_value(kc2,kc1)
     data_tmp_b.append(tt*v_orth*kappa)
     data_tmp_b2.append(tt)
     if tt == 1.0:
      kappa_mask.append([kc1,kc2])
     
     #---------------------
     B[kc1] += self.mesh.get_side_periodic_value(kc2,kc1)*v_orth/vol1*kappa

    #Boundary_elements
    FF = np.zeros(self.n_elems)
    boundary_mask = np.zeros(self.n_elems)
    for side in self.mesh.side_list['Boundary']:
     elem = self.mesh.side_elem_map[side][0]
     vol = self.mesh.get_elem_volume(elem)
     area = self.mesh.get_side_area(side)
     FF[elem] = area/vol
     boundary_mask[elem] = 1.0

    F = csc_matrix((np.array(data_tmp),(np.array(row_tmp),np.array(col_tmp))), shape=(self.n_elems,self.n_elems) )
    RHS = csc_matrix( (np.array(data_tmp_b),(np.array(row_tmp_b),np.array(col_tmp_b))), shape=(self.n_elems,self.n_elems) )
    PER = csc_matrix( (np.array(data_tmp_b2),(np.array(row_tmp_b),np.array(col_tmp_b))), shape=(self.n_elems,self.n_elems) )

    #-------
    data = {'F':F,'RHS':RHS,'B':B,'PER':PER,'boundary_mask':boundary_mask,'FF':FF,'kappa_mask':kappa_mask}
   else: data = None
   data =  MPI.COMM_WORLD.bcast(data,root=0)
   self.F = data['F']
   self.RHS = data['RHS']
   self.PER = data['PER']
   self.kappa_mask = data['kappa_mask']
   self.B = data['B']
   self.FF = data['FF']
   self.boundary_mask = data['boundary_mask']


  def solve_fourier(self,**argv):
       
    rank = MPI.COMM_WORLD.rank    
    #Update matrices---------------------------------------------------
    G = argv.setdefault('interface_conductance',np.zeros(1))#*self.dom['correction']
    TL = argv.setdefault('lattice_temperature',np.zeros((1,self.n_el)))
    TB = argv.setdefault('boundary_temperature',np.zeros((1,self.n_el)))
    mfe_factor = argv.setdefault('mfe_factor',0.0)
    kappa_bulk = argv.setdefault('kappa',[1.0])
   
    
    
    #print(kappa_bulk)
    #G *=5.0
    n_index = len(kappa_bulk)

    #G = np.zeros(n_index)
    #-------------------------------
    KAPPAp,KAPPA = np.zeros((2,n_index))
    FLUX,FLUXp = np.zeros((2,n_index,3,self.n_elems))
    TLp,TLtot = np.zeros((2,n_index,self.n_elems))

    comm = MPI.COMM_WORLD  
    size = comm.size
    rank = comm.rank
    block = n_index // size + 1
    for kk in range(block):
     index = rank*block + kk   
     if index < n_index:
    
    #set up MPI------------------
    #for index in range(n_index):
      #Aseemble matrix-------------       
     
      A  = kappa_bulk[index] * self.F + mfe_factor * csc_matrix(scipy.sparse.eye(self.n_elems)) 
      A += mfe_factor * diags([self.FF],[0]).tocsc() * G[index]  
      
      B  = kappa_bulk[index] * self.B + mfe_factor * TL[index] 
      B += mfe_factor * np.multiply(self.FF,TB[index]) * G[index]
      
      if mfe_factor == 0.0: A[10,10] = 1.0
      SU = splu(A)
      C = np.zeros(self.n_el)
      #--------------------------------------

      n_iter = 0
      kappa_old = 0
      error = 1        

      
      while error > argv.setdefault('max_fourier_error',1e-2) and \
                    n_iter < argv.setdefault('max_fourier_iter',10) :
                    
        RHS = B + C*kappa_bulk[index]
        if mfe_factor == 0.0: RHS[10] = 0.0      
        temp = SU.solve(RHS)
        temp = temp - (max(temp)+min(temp))/2.0
        (C,flux) = self.compute_non_orth_contribution(temp)
        
        KAPPAp[index] = self.compute_diffusive_thermal_conductivity(temp)*kappa_bulk[index]
        error = abs((KAPPAp[index] - kappa_old)/KAPPAp[index])
        kappa_old = KAPPAp[index]
        n_iter +=1
        
      #FLUXp = np.array([-self.elem_kappa_map[k]*tmp for k,tmp in enumerate(flux)])
      FLUXp[index] = flux.T
      TLp[index] = temp
     
        
    comm.Allreduce([KAPPAp,MPI.DOUBLE],[KAPPA,MPI.DOUBLE],op=MPI.SUM)
    comm.Allreduce([TLp,MPI.DOUBLE],[TLtot,MPI.DOUBLE],op=MPI.SUM) #to be changed
    comm.Allreduce([FLUXp,MPI.DOUBLE],[FLUX,MPI.DOUBLE],op=MPI.SUM)  
      
    
    
    if argv.setdefault('verbose',True) and rank==0:

     print('  ')
     print('Thermal Conductivity: '.ljust(20) +  '{:8.2f}'.format(KAPPAp.sum())+ ' W/m/K')
     print('  ')
    


    return {'kappa_fourier':KAPPA,'temperature_fourier_gradient':FLUX,'temperature_fourier':TLtot}

      
  def compute_non_orth_contribution(self,temp) :

    self.gradT = self.mesh.compute_grad(temp)
    C = np.zeros(self.n_el)
    for i,j in zip(*self.mesh.A.nonzero()):

     if not i==j:
      #ll = self.mesh.get_side_between_two_elements(i,j)
      #kappa = self.side_kappa_map[ll]
      #Get agerage gradient----
      side = self.mesh.get_side_between_two_elements(i,j)


      w = self.mesh.get_interpolation_weigths(side,i)
      grad_ave = w*self.gradT[i] + (1.0-w)*self.gradT[j]
      #------------------------
      (dumm,v_non_orth) = self.mesh.get_decomposed_directions(i,j)

      
      C[i] += np.dot(grad_ave,v_non_orth)/2.0/self.mesh.get_elem_volume(i)#*kappa
      C[j] -= np.dot(grad_ave,v_non_orth)/2.0/self.mesh.get_elem_volume(j)#*kappa


    return C, self.gradT# [-self.elem_kappa_map[n]*tmp for n,tmp in enumerate(self.gradT)]


  def compute_inho_diffusive_thermal_conductivity(self,temp,mat=np.eye(3)):

   kappa = 0
   for i,j in zip(*self.PER.nonzero()):

    #ll = self.mesh.get_side_between_two_elements(i,j)
    #kappa_b = self.side_kappa_map[ll]
    (v_orth,dummy) = self.mesh.get_decomposed_directions(i,j,rot=mat)
   
    
    kappa += 0.5*v_orth *self.PER[i,j]*(temp[j]+self.PER[i,j]-temp[i])#*kappa_b

   return kappa*self.kappa_factor


  def compute_diffusive_thermal_conductivity(self,temp,mat=np.eye(3)):

   
   kappa = 0
   for (i,j) in self.kappa_mask:
    (v_orth,dummy) = self.mesh.get_decomposed_directions(i,j,rot=mat)
    kappa += v_orth  *(temp[j]+1.0-temp[i])

   #kappa = 0
   #for i,j in zip(*self.PER.nonzero()):
   # (v_orth,dummy) = self.mesh.get_decomposed_directions(i,j,rot=mat)
   # kappa += 0.5*v_orth *self.PER[i,j]*(temp[j]+self.PER[i,j]-temp[i])

   return kappa*self.kappa_factor

  def print_logo(self):

   if MPI.COMM_WORLD.Get_rank() == 0:
    print(' ')
    print(r'''  ___                   ____ _____ _____ ''' )
    print(r''' / _ \ _ __   ___ _ __ | __ )_   _| ____|''')
    print(r'''| | | | '_ \ / _ \ '_ \|  _ \ | | |  _|  ''')
    print(r'''| |_| | |_) |  __/ | | | |_) || | | |___ ''')
    print(r''' \___/| .__/ \___|_| |_|____/ |_| |_____|''')
    print(r'''      |_|                                ''')
    print('')
    print('Giuseppe Romano [romanog@mit.edu]')
    print(' ')

  def print_dof(self):

   if MPI.COMM_WORLD.Get_rank() == 0:
    if self.verbose:
     print(' ')
     print('Elements:          ' + str(self.n_el))
     print('Azimuthal angles:  ' + str(self.n_theta))
     #print('Polar angles:      ' + str(self.n_phi))
     print('Reduced Mean-free-paths:   ' + str(self.n_rmfp))
     #print('Degree-of-freedom: ' + str(self.n_mfp*self.n_theta*self.n_phi*self.n_el))
     print('Degree-of-freedom: ' + str(self.n_rmfp*self.n_theta*self.n_phi*self.n_el))


      #Thermalize boundaries--------------------------
      #Hot = np.zeros(self.n_elems)
      #for side in self.mesh.side_list['Hot']:
   #    (coeff,elem) = self.mesh.get_side_coeff(side)#
   #    tmp = np.dot(coeff,angle_factor)#
   #    if tmp < 0:
   #     Hot[elem] = -tmp*0.5#
   #    else:
   #     r.append(elem); c.append(elem); d.append(tmp)

    #  Cold = np.zeros(self.n_elems)
    #  for side in self.mesh.side_list['Cold']:#
       #(coeff,elem) = self.mesh.get_side_coeff(side)
       #tmp = np.dot(coeff,angle_factor)
       #if tmp < 0:
       # Cold[elem] = tmp*0.5
       #else:
       # r.append(elem); c.append(elem); d.append(tmp)

      #-----------------------------------------------
  #Hot.dump(open(self.cache + '/Hot_' + str(index) +r'.np','w+'))
  #Cold.dump(open(self.cache +'/Cold_' + str(index) +r'.np','w+'))



from __future__ import print_function
from __future__ import absolute_import
import os,sys
import numpy as np
import subprocess
from mpi4py import MPI
from pyvtk import *
from .GenerateSquareLatticePores import *
from .GenerateHexagonalLatticePores import *
from .GenerateStaggeredLatticePores import *
from .GenerateCustomPores import *
from .GenerateRandomPoresOverlap import *
from .GenerateRandomPoresGrid import *
from . import GenerateMesh2D
from . import GenerateBulk2D
from . import GenerateBulk3D
from . import porous
import matplotlib
if not matplotlib.get_backend() == 'Agg': matplotlib.use('Agg')
import matplotlib.patches as patches
from .fig_maker import *
from matplotlib.path import Path
from IPython.display import display, HTML
from shapely.geometry import LineString
import shapely
import pickle
import sparse

#import GenerateInterface2D
#from nanowire import *
import deepdish as dd
from scipy.sparse import csc_matrix
from matplotlib.pylab import *
from shapely.geometry import MultiPoint,Point,Polygon,LineString


def create_path(obj):

   codes = [Path.MOVETO]
   for n in list(range(len(obj)-1)):
    codes.append(Path.LINETO)
   codes.append(Path.CLOSEPOLY)

   verts = []
   for tmp in obj:
     verts.append(tmp)
   verts.append(verts[0])

   path = Path(verts,codes)
   return path

class Geometry(object):

 def __init__(self,**argv):

  direction = argv.setdefault('direction','x')
  if direction == 'x':self.direction = 0
  if direction == 'y':self.direction = 1
  if direction == 'z':self.direction = 2
  self.argv = argv

  geo_type = argv['type']

  if geo_type == 'load':
  # if MPI.COMM_WORLD.Get_rank() == 0:
    self.state = pickle.load(open('geometry.p','rb'))
      
    self._update_data()

  else:
   if MPI.COMM_WORLD.Get_rank() == 0:
    #porous-----
    #if geo_type == 'porous/random' or \
    polygons = []

    if  geo_type == 'porous/square_lattice' or\
        geo_type == 'porous/hexagonal_lattice' or\
        geo_type == 'porous/staggered_lattice' or\
        geo_type == 'porous/random_over_grid' or\
        geo_type == 'porous/random' or\
        geo_type == 'porous/custom':

     if geo_type == 'porous/square_lattice':
      frame,polygons = GenerateSquareLatticePores(argv)

     if geo_type == 'porous/hexagonal_lattice':
      frame,polygons = GenerateHexagonalLatticePores(argv)

     if geo_type == 'porous/staggered_lattice':
      frame,polygons = GenerateStaggeredLatticePores(argv)

     if geo_type == 'porous/custom':
      frame,polygons = GenerateCustomPores(argv)

     if geo_type == 'porous/random':
      frame,polygons = GenerateRandomPoresOverlap(argv)

     if geo_type == 'porous/random_over_grid':
      x,polygons = GenerateRandomPoresGrid(**argv)
      argv['polygons'] = polygons
      frame,polygons = GenerateCustomPores(argv)


    self.Lz = float(argv.setdefault('lz',0.0))
    if geo_type == 'bulk':
      Lx = float(argv['lx'])
      Ly = float(argv['ly'])

      frame = []
      frame.append([-Lx/2,Ly/2])
      frame.append([Lx/2,Ly/2])
      frame.append([Lx/2,-Ly/2])
      frame.append([-Lx/2,-Ly/2])

    self.frame = frame
    self.polygons = polygons

    if argv.setdefault('mesh',True):
     self.mesh(**argv)

   MPI.COMM_WORLD.Barrier()
  if argv.setdefault('save_fig',False):
   self.plot_polygons()


 def mesh(self,**argv):

     
    if self.Lz > 0.0:
     self.dim = 3
    else:
     self.dim = 2

    
    argv.update({'lz':self.Lz})
    if len(self.polygons) > 0 and self.dim == 3:
     #GenerateMesh3D.mesh(self.polygons,self.frame,argv)
     argv['polygons'] = self.polygons
     argv['frame'] = self.frame
     porous.Porous(**argv)

    if len(self.polygons) > 0 and self.dim == 2:
     GenerateMesh2D.mesh(self.polygons,self.frame,argv)
    if len(self.polygons) == 0 and self.dim == 2:
     GenerateBulk2D.mesh(argv)
    if len(self.polygons) == 0 and self.dim == 3:
     GenerateBulk3D.mesh(argv)
    #if argv['model'] == 'nanowire':
    # nanowire(argv)
    # self.dim = 3
    #Interface----EXPERIMENTAL
    #if argv['model'] == '2DInterface':
    #  Generate:/2D.mesh(argv)
    #  self.dim = 2

    #bulk----------
    #if geo_type == 'bulk':
    # if 'lz' in argv.keys() == 3:
    #  GenerateBulk3D.mesh(argv)
    #  self.dim = 3
    # else:
    #  GenerateBulk2D.mesh(argv)
    #  self.dim = 2
     #-----------------------------------
    data = self.compute_mesh_data()
    
        
    #dd.io.save('geometry.hdf5', data)
    pickle.dump(data,open('geometry.p','wb'),protocol=pickle.HIGHEST_PROTOCOL)
    
 # MPI.COMM_WORLD.Barrier()

 def get_repeated_size(self,argv):
    Nx = argv.setdefault('repeat_x',1)
    Ny = argv.setdefault('repeat_y',1)
    size = self.size
    Lx = size[0]*Nx
    Ly = size[1]*Ny
    return Lx,Ly


 def get_interface_point_couples(self,argv):

   Nx = argv.setdefault('repeat_x',1)
   Ny = argv.setdefault('repeat_y',1)


   for ll in self.side_list['Interface']:
    p1 = self.nodes[self.sides[ll][0]][:2]
    p2 = self.nodes[self.sides[ll][1]][:2]
    for nx in range(Nx):
      for ny in range(Ny):
       P = np.array([self.size[0]*(nx-(Nx-1)*0.5),\
                     self.size[1]*(ny-(Ny-1)*0.5)])
       p1 = np.array(p1) + P
       p2 = np.array(p2) + P
       plot([p1[0],p2[0]],[p1[1],p2[1]],color='w',ls='--',zorder=1)




 def plot_polygons(self,**argv):

   if MPI.COMM_WORLD.Get_rank() == 0:

    lx = abs(self.frame[0][0])*2
    ly = abs(self.frame[0][1])*2

    #init_plotting()
    close()
    fig = figure(num=" ", figsize=(8*lx/ly, 4), dpi=80, facecolor='w', edgecolor='k')
    #axes([0,0,1.0,1.0])
    axes([0,0,0.5,1.0])


    xlim([-lx/2.0,lx/2.0])
    ylim([-ly/2.0,ly/2.0])

    path = create_path(self.frame)
    patch = patches.PathPatch(path,linestyle=None,linewidth=0.1,color='gray',zorder=1,joinstyle='miter')
    gca().add_patch(patch);

    if self.argv.setdefault('inclusion',False):
     color='g'
    else:
     color='white'

    for poly in self.polygons:
     path = create_path(poly)
     patch = patches.PathPatch(path,linestyle=None,linewidth=0.1,color=color,zorder=2,joinstyle='miter')
     gca().add_patch(patch);
     
     
    #plot Boundary Conditions-----
    if argv.setdefault('plot_boundary',False):
     for side in self.side_list['Boundary'] + self.side_list['Interface'] :
      p1 = self.sides[side][0]
      p2 = self.sides[side][1]
      n1 = self.nodes[p1]
      n2 = self.nodes[p2]
      plot([n1[0],n2[0]],[n1[1],n2[1]],color='#f77f0e',lw=6)
     
     
      #plot Periodic Conditions-----
    if argv.setdefault('plot_boundary',False):
     for side in self.side_list['Periodic'] + self.side_list['Inactive']  :
        
      p1 = self.sides[side][0]
      p2 = self.sides[side][1]
      n1 = self.nodes[p1]
      n2 = self.nodes[p2]
      plot([n1[0],n2[0]],[n1[1],n2[1]],color='#1f77b4',lw=12,zorder=3)

       
    
    #----------------------------
     
     
    axis('off')

    #if selfargv.setdefault('show',False):
     #print(fig.canvas.toolbar)
     #quit()
    #show()
     #rcParams['toolbar']='None'

    #fig:wq

    savefig(argv.setdefault('fig_file','geometry.png'))

    #return fig




 def compute_triangle_area(self,p1,p2,p):
   p = Polygon([(p1[0],p1[1]),(p2[0],p2[1]),(p[0],p[1])])
   return p.area


 def compute_2D_interpolation(self,data,p,elem):

   nodes = self.elems[elem]
   p1 = self.nodes[nodes[0]]
   p2 = self.nodes[nodes[1]]
   p3 = self.nodes[nodes[2]]

   v1 = data[nodes[0]]
   v2 = data[nodes[1]]
   v3 = data[nodes[2]]

   area_1 = self.compute_triangle_area(p2,p3,p)
   area_2 = self.compute_triangle_area(p1,p3,p)
   area_3 = self.compute_triangle_area(p1,p2,p)
   area = area_1 + area_2 + area_3
   v = v1*area_1 + v2*area_2 + v3*area_3
   v /= area

   return v




 def adjust_boundary_elements(self):

  for side in self.side_list['Boundary']:
   self.side_elem_map[side].append(self.side_elem_map[side][0])

 def compute_mesh_data(self):


    self.import_mesh()

    self.compute_elem_map()
    self.adjust_boundary_elements()
    self.compute_elem_volumes()
    self.compute_side_areas()
    self.compute_side_normals()
    self.compute_elem_centroids()
    self.compute_side_centroids()
    self.compute_least_square_weigths()
    self.compute_connecting_matrix()
    self.compute_connecting_matrix_new()
    self.compute_interpolation_weigths()
    self.compute_contact_areas()
    self.compute_boundary_condition_data()


    data = {'side_list':self.side_list,\
          'node_list':self.node_list,\
          'exlude':self.exlude,\
          'elem_side_map':self.elem_side_map,\
          'elem_region_map':self.elem_region_map,\
          'side_elem_map':self.side_elem_map,\
          'side_node_map':self.side_node_map,\
          'node_elem_map':self.node_elem_map,\
          'elem_map':self.elem_map,\
          'nodes':self.nodes,\
          'N':self.N,\
          'sides':self.sides,\
          'elems':self.elems,\
          'A':self.A,\
          'side_periodicity':self.side_periodicity,\
          'dim':self.dim,\
          'weigths':self.weigths,\
          'region_elem_map':self.region_elem_map,\
          'size':self.size,\
          'c_areas':self.c_areas,\
          'interp_weigths':self.interp_weigths,\
          'elem_centroids':self.elem_centroids,\
          'side_centroids':self.side_centroids,\
          'elem_volumes':self.elem_volumes,\
          'periodic_nodes':self.periodic_nodes,\
          'side_areas':self.side_areas,\
          'pairs':self.pairs,\
          'B':self.B,\
          'CM':self.CM,\
          'CP':self.CP,\
          'CPB':self.CPB,\
          'boundary_elements':self.boundary_elements,\
          'interface_elements':self.interface_elements,\
          'side_normals':self.side_normals,\
          'grad_direction':self.direction,\
          'area_flux':self.area_flux,\
          'flux_sides':self.flux_sides,\
          'frame':self.frame,\
          'polygons':self.polygons,\
          'argv':self.argv,\
          'side_periodic_value':self.side_periodic_value}

    return data


 def compute_connecting_matrix(self):

   nc = len(self.elems)
   row_tmp = []
   col_tmp = []
   data_tmp = []

   for ll in self.side_list['active'] :
    if not ll in self.side_list['Hot'] and\
     not ll in self.side_list['Cold']:

         
     elems = self.get_elems_from_side(ll)
     kc1 = elems[0]
     kc2 = elems[1]
     row_tmp.append(kc1)
     col_tmp.append(kc2)
     data_tmp.append(1)
     row_tmp.append(kc2)
     col_tmp.append(kc1)
     data_tmp.append(1)
   
   self.A = csc_matrix( (np.array(data_tmp),(np.array(row_tmp),np.array(col_tmp))), shape=(nc,nc) ).todense()

 def compute_connecting_matrix_new(self):

   nc = len(self.elems)
   N = sparse.DOK((nc,nc,3), dtype=float32)
   self.CM = np.zeros((nc,3))
   self.CP = np.zeros((nc,3))
   self.CPB = np.zeros((nc,3))
   i = [];j = [];k = []
   data = []
   for ll in self.side_list['active'] :
     elems = self.get_elems_from_side(ll)
     kc1 = elems[0]
     kc2 = elems[1]
     vol1 = self.get_elem_volume(kc1)
     vol2 = self.get_elem_volume(kc2)
     area = self.compute_side_area(ll)
     normal = self.compute_side_normal(kc1,ll)
     if not kc1 == kc2:
       N[kc1,kc2] = normal*area/vol1
       N[kc2,kc1] = -normal*area/vol2
     else:  
       self.CM[kc1]  += normal*area/vol1
       self.CPB[kc1] += normal*area/vol1
       self.CP[kc1]  += normal


   self.N = N.to_coo()




 def compute_elem_map(self):

  self.elem_map = {}
  for elem1 in self.elem_side_map:
    for side in self.elem_side_map[elem1]:
     elem2 = self.get_neighbor_elem(elem1,side)
     self.elem_map.setdefault(elem1,[]).append(elem2)


 def get_side_orthognal_direction(self,side):

   elem = self.side_elem_map[side][0]
   c1 = self.get_elem_centroid(elem)
   c2 = self.get_side_centroid(side)
   normal = self.compute_side_normal(elem,side)
   area = self.compute_side_area(side)
   Af = area*normal
   dist = c2-c1
   v_orth = np.dot(Af,Af)/np.dot(Af,dist)
   return v_orth


 def get_distance_between_centroids_of_two_elements_from_side(self,ll):


   (elem_1,elem_2) = self.side_elem_map[ll]
   c1 = self.get_elem_centroid(elem_1)
   c2 = self.get_next_elem_centroid(elem_1,ll)
   dist = np.linalg.norm(c2-c1)

   return dist

 def get_decomposed_directions(self,elem_1,elem_2,rot = np.eye(3)):

   if elem_1 == elem_2:
    return 0.0,0.0
   else:
    side = self.get_side_between_two_elements(elem_1,elem_2)
    normal = self.compute_side_normal(elem_1,side)
    area = self.compute_side_area(side)
    Af = area*normal
    c1 = self.get_elem_centroid(elem_1)
    c2 = self.get_next_elem_centroid(elem_1,side)
    dist = c2-c1
    v_orth = area*np.dot(normal,np.dot(rot,normal))/np.dot(normal,dist)
    v_non_orth = np.dot(rot,Af) - dist*v_orth
    return v_orth,v_non_orth


 def get_side_coeff(self,side):

  elem = self.side_elem_map[side][0]
  vol = self.get_elem_volume(elem)
  normal = self.compute_side_normal(elem,side)
  area = self.get_side_area(side)
  return normal*area/vol #This is specific to boundaries

 #def get_coeff(self,elem_1,elem_2):

 #  side = self.get_side_between_two_elements(elem_1,elem_2)
 #  vol = self.get_elem_volume(elem_1)
 #  normal = self.compute_side_normal(elem_1,side)
 #  area = self.compute_side_area(side)

 #  return normal*area/vol

 def compute_2D_adjusment(self,normal,phi_s,dphi):

  tmp =  np.arctan2(normal[1],normal[0])
  if not tmp == 0.0:
   tmp = (tmp + np.pi*(1.0-np.sign(tmp)))%(2.0*np.pi)

  phi_normal = (np.pi/2.0-tmp)%(2.0*np.pi)

  phi_right = (phi_normal+np.pi/2.0)%(2.0*np.pi)
  phi_left = (phi_normal-np.pi/2.0)%(2.0*np.pi)

  phi_1 = 0.0
  phi_2 = 0.0
  phi_left = round(phi_left,8)
  phi_right = round(phi_right,8)
  phi_normal = round(phi_normal,8)
  phi_s = round(phi_s,8)


  delta = 1e-7
  case = 0
  if phi_s <= phi_right +delta and phi_s > (phi_right - dphi/2.0)%(2.0*np.pi):
   phi_1 = phi_right
   phi_2 = (phi_s + dphi/2.0)%(2.0*np.pi)
   case = 1

  elif phi_s > phi_right + delta and phi_s < (phi_right + dphi/2.0)%(2.0*np.pi):
   phi_1 = (phi_s-dphi/2.0)%(2.0*np.pi)
   phi_2 = phi_right
   case = 2

  elif phi_s < phi_left - delta and phi_s > (phi_left - dphi/2.0)%(2.0*np.pi):
   phi_1 = phi_left
   phi_2 = (phi_s + dphi/2.0)%(2.0*np.pi)
   case = 3
   #print(phi_1 * 180.0/np.pi)
   #print(phi_2 * 180.0/np.pi)
   #print(phi_normal * 180.0/np.pi)
   #print(phi_s * 180.0/np.pi)



  elif phi_s >= phi_left - delta and phi_s < (phi_left + dphi/2.0)%(2.0*np.pi):
   phi_1 = (phi_s-dphi/2.0)%(2.0*np.pi)
   phi_2 = phi_left
   case = 4


  phi_dir_adj = [np.cos(phi_1) - np.cos(phi_2),np.sin(phi_2) - np.sin(phi_1),0.0]
  return np.array(phi_dir_adj),case


 def get_angular_coeff_old(self,elem_1,elem_2,angle):

  vol = self.get_elem_volume(elem_1)
  side = self.get_side_between_two_elements(elem_1,elem_2)
  normal = self.compute_side_normal(elem_1,side)
  area = self.compute_side_area(side)
  tmp = np.dot(angle,normal)
  cm = 0
  cp = 0
  cmb = 0
  cpb = 0

   
  if tmp < 0:
   if elem_1 == elem_2:  
    cmb = tmp*area/vol
   else:
    cm = tmp*area/vol
  else:
   cp = tmp*area/vol
   if elem_1 == elem_2:  
    cpb = tmp
  

  return cm,cp,cmb,cpb


 def get_angular_coeff(self,elem_1,elem_2,index):

  side = self.get_side_between_two_elements(elem_1,elem_2)
  vol = self.get_elem_volume(elem_1)
  normal = self.compute_side_normal(elem_1,side)
  for i in range(3):
   normal[i] = round(normal[i],8)
  area = self.compute_side_area(side)


  if self.dim == 2:
      
   p = index
   
   polar_int = self.dom['polar_dir'][p]*self.dom['polar_factor']
   tmp = np.dot(polar_int,normal)
   coeff  = tmp*area/vol
   #coeff = np.dot(angle_factor,normal)*area/vol
   #control  = np.dot(self.dom['polar_dir'][p],normal)
  else :
   t = int(index/self.dom['self.n_phi'])
   p = index%self.dom['n_phi']
   angle_factor = self.dom['S'][t][p]/self.dom['d_omega'][t][p]
   coeff = np.dot(angle_factor,normal)*area/vol

  #coeff = round(coeff,8)
  #anti aliasing---
  extra_coeff = 0.0
  extra_angle = 0.0
  case = 0
  if self.dim == 2:
   (phi_dir_adj,case) = self.compute_2D_adjusment(normal,round(self.dom['phi_vec'][p],8),self.dom['d_phi'])
   #extra_coeff = np.dot(phi_dir_adj,normal)*area/vol/self.dom['d_phi_vec'][p]
   #extra_angle = np.dot(phi_dir_adj,normal)*self.dom['d_phi']

  cm = 0.0
  cp = 0.0
  cmb = 0.0
  cpb = 0.0
  if coeff >= 0:
   cp = coeff - extra_coeff
   if not elem_1 == elem_2:
    cm = extra_coeff
   else:
    cpb = tmp*self.dom['d_phi'] - extra_angle
    cmb = extra_coeff
  else :
   cp = extra_coeff
   if not elem_1 == elem_2: #In this case the incoming part is from the boundary
    cm = coeff  - extra_coeff
   else:
    cmb = coeff - extra_coeff
    cpb = extra_angle


  if cpb < 0.0:
   print('cpb')
   print(cpb)

  if cmb > 0.0:
   print('cmb')
   print(cmb)

  if cm > 0.0:
   print('cm')
   print(cm)

  if cp < 0.0:#
     print('cp')
     print(cp)
     print(extra_coeff)
     print(case)
     print(coeff)
#     quit()

  return cm,cp,cmb,cpb


# def get_aa(self,elem_1,elem_2,phi_dir,dphi):

#   side = self.get_side_between_two_elements(elem_1,elem_2)
#   vol = self.get_elem_volume(elem_1)
#   normal = self.compute_side_normal(elem_1,side)

#   gamma = arcsin(np.dot(phi_dir,normal[0:2]))

#   if gamma >= 0.0:
     #p_plus = 0.5 + min([gamma,dphi/2.0])/dphi
     #p_minus = 1.0-p_plus


   #if gamma < 0.0:
#     p_minus = 0.5 + min([-gamma,dphi/2.0])/dphi
     #p_plus = 1.0-p_minus

  # return p_minus,p_plus



 def get_af(self,elem_1,elem_2):

   side = self.get_side_between_two_elements(elem_1,elem_2)
   normal = self.compute_side_normal(elem_1,side)
   area = self.compute_side_area(side)
   return normal*area


 def get_normal_between_elems(self,elem_1,elem_2):

   side = self.get_side_between_two_elements(elem_1,elem_2)
   normal = self.compute_side_normal(elem_1,side)
   return normal


 def get_side_between_two_elements(self,elem_1,elem_2):

   if elem_1 == elem_2: #Boundary side
    

    for side in self.elem_side_map[elem_1]:
     if side in self.side_list['Boundary']:
      return side
    print('No boundary side')
    quit()
   else:

    for side_1 in self.elem_side_map[elem_1]:
     for side_2 in self.elem_side_map[elem_2]:
      if side_1 == side_2:
       return side_1

    print('no adjacents elems')
    assert(1==0)
    quit()



 def compute_gradient_on_side(self,x,ll,side_periodic_value):

   #Diff in the temp
  
   elems = self.side_elem_map[ll]
   kc1 = elems[0]
   temp_1 = x[kc1]
   #Build tempreature matrix--------------------------
   diff_temp = np.zeros(self.dim + 1)
   grad = np.zeros(3)
   for s in self.elem_side_map[kc1]:
    if s in self.side_list['Boundary']:
     temp_2 = temp_1
    else:
     kc2 = self.get_neighbor_elem(kc1,s)
     temp_2 = x[kc2] + self.get_side_periodic_value(s,kc2,side_periodic_value)

    ind1 = self.elem_side_map[kc1].index(s)
    diff_t = temp_2 - temp_1
    diff_temp[ind1] = diff_t
    #-----------------------------------------------
    #COMPUTE GRADIENT
   tmp = np.dot(self.weigths[kc1],diff_temp)
   grad[0] = tmp[0] #THESE HAS TO BE POSITIVE
   grad[1] = tmp[1]
   if self.dim == 3:
    grad[2] = 0.0

   return grad



 def compute_thermal_conductivity_from_scalar(self,x,flux_sides,side_periodic_value,factor = np.eye(3)):

   #Diff in the temp
   kappa = 0.0
   for ll in flux_sides :
    normal = self.get_side_normal(0,ll)
    area_loc = self.get_side_area(ll)
    elems = self.side_elem_map[ll]
    kc1 = elems[0]
    temp_1 = x[kc1]
    #Build tempreature matrix--------------------------
    diff_temp = np.zeros(self.dim + 1)
    grad = np.zeros(self.dim)
    for s in self.elem_side_map[kc1]:
     if s in self.side_list['Boundary']:
      temp_2 = temp_1
     else:
      kc2 = self.get_neighbor_elem(kc1,s)
      temp_2 = x[kc2] + self.get_side_periodic_value(s,kc2,side_periodic_value)

     ind1 = self.elem_side_map[kc1].index(s)
     diff_t = temp_2 - temp_1
     diff_temp[ind1] = diff_t
    #-----------------------------------------------
    #COMPUTE GRADIENT
    tmp = np.dot(self.weigths[kc1],diff_temp)
    grad[0] = tmp[0] #THESE HAS TO BE POSITIVE
    grad[1] = tmp[1]
    if self.dim == 3:
     grad[2] = tmp[2]

    #COMPUTE THERMAL CONDUCTIVITY------------------
    for i in range(self.dim):
     for j in range(self.dim):
      kappa -= normal[i]*factor[i][j]*grad[j]*area_loc

   return kappa


 def compute_thermal_conductivity_outer(self,vector,scalar,flux_sides):

   #Here we simply invert Fourier's law
   factor = np.eye(3)
   power = 0.0
   for ns in flux_sides:
    ne = self.side_elem_map[ns][0]
    normal = self.get_side_normal(0,ns)
    area_loc = self.get_side_area(ns)
    power_loc = 0.0
    for i in range(self.dim):
     for j in range(self.dim):
      power_loc -= normal[i]*factor[i][j]*vector[j]*scalar[ne]*area_loc
    power += power_loc
   return power

 def compute_thermal_conductivity(self,grad,flux_sides,factor=np.eye(3)):

   #Here we simply invert Fourier's law
   power = 0.0
   for ns in flux_sides:
    ne = self.side_elem_map[ns][0]
    normal = self.get_side_normal(0,ns)
    area_loc = self.get_side_area(ns)
    power_loc = 0.0
    for i in range(3):
     for j in range(3):
      power_loc -= normal[i]* factor[i][j] * grad[ne][j]*area_loc
    power += power_loc

   return power



 def compute_average_temperature(self,x,flux_sides):

   #Here we simply invert Fourier's law
   temp = 0.0
   for ns in flux_sides:
    ne = self.side_elem_map[ns][0]
    area_loc = self.get_side_area(ns)
    temp += area_loc * x[ne]

   return temp


 def compute_contact_areas(self):

  self.c_areas = np.zeros(3)

  nodesT = self.nodes.copy().T

  minx = min(nodesT[0])
  maxx = max(nodesT[0])
  miny = min(nodesT[1])
  maxy = max(nodesT[1])
  minz = min(nodesT[2])
  maxz = max(nodesT[2])
  if self.dim == 3:
   self.c_areas[0] = (maxy - miny) * (maxz - minz)
   self.c_areas[1] = (maxz - minz) * (maxx - minx)
   self.c_areas[2] = (maxy - miny) * (maxx - minx)
  else :
   self.c_areas[0] = maxy - miny
   self.c_areas[1] = maxx - minx
   self.c_areas[2] = 0


 def get_side_area(self,side):
  return self.side_areas[side]

 def get_elem_volume(self,elem):
  return self.elem_volumes[elem]

 def compute_elem_volumes(self):

  self.elem_volumes = np.zeros(len(self.elems))
  for k in range(len(self.elems)):
   self.elem_volumes[k] = self.compute_elem_volume(k)




 def compute_side_areas(self):
  self.side_areas = np.zeros(len(self.sides))
  for k in range(len(self.sides)):
   self.side_areas[k] = self.compute_side_area(k)


 def get_dim(self):
   return self.dim

 def _update_data(self):
    self.nodes = self.state['nodes']
    self.dim = self.state['dim']
    self.A = self.state['A']
    self.sides = self.state['sides']
    self.elems = self.state['elems']
    self.side_periodicity = self.state['side_periodicity']
    self.region_elem_map = self.state['region_elem_map']
    self.size = self.state['size']
    self.dim = self.state['dim']
    self.weigths = self.state['weigths']
    self.elem_region_map = self.state['elem_region_map']
    self.elem_side_map = self.state['elem_side_map']
    self.side_node_map = self.state['side_node_map']
    self.node_elem_map = self.state['node_elem_map']
    self.side_elem_map = self.state['side_elem_map']
    self.side_list = self.state['side_list']
    self.interp_weigths = self.state['interp_weigths']
    self.elem_centroids = self.state['elem_centroids']
    self.c_areas = self.state['c_areas']
    self.side_centroids = self.state['side_centroids']
    self.periodic_nodes = self.state['periodic_nodes']
    self.elem_volumes = self.state['elem_volumes']
    self.boundary_elements = self.state['boundary_elements']
    self.interface_elements = self.state['interface_elements']
    self.side_normals = self.state['side_normals']
    self.side_areas = self.state['side_areas']
    self.exlude = self.state['exlude']
    self.pairs = self.state['pairs']
    self.direction = self.state['grad_direction']
    self.area_flux = self.state['area_flux']
    self.node_list = self.state['node_list']
    self.flux_sides = self.state['flux_sides']
    self.frame = self.state['frame']
    self.argv = self.state['argv']
    self.polygons = self.state['polygons']
    self.N = self.state['N']
    self.CM = self.state['CM']
    self.CP = self.state['CP']
    self.CPB = self.state['CPB']
    self.B = self.state['B']
    self.side_periodic_value = self.state['side_periodic_value']
    self.kappa_factor = self.size[self.direction]/self.area_flux



 def import_mesh(self):

  #Create mesh---
  subprocess.check_output(['gmsh','-' + str(self.dim),'mesh.geo','-o','mesh.msh'])

  #-------------

  f = open('mesh.msh','r')
  #Read boundary conditions------------------
  f.readline()
  f.readline()
  f.readline()
  f.readline()
  self.blabels = {}
  nb =  int(f.readline())
  for n in range(nb):
   tmp = f.readline().split()
   l = tmp[2].replace('"',r'')
   self.blabels.update({int(tmp[1]):l})


  #------------------------------------------
  self.elem_region_map = {}
  self.region_elem_map = {}


  #import nodes------------------------
  #f.readline()
  f.readline()
  f.readline()
  n_nodes = int(f.readline())
  nodes = []
  for n in range(n_nodes):
   tmp = f.readline().split()
   nodes.append([float(tmp[1]),float(tmp[2]),float(tmp[3])])

  nodes = np.array(nodes)

  if self.dim == 3:
   self.size = np.array([ max(nodes[:,0]) - min(nodes[:,0]),\
                  max(nodes[:,1]) - min(nodes[:,1]),\
                  max(nodes[:,2]) - min(nodes[:,2])])
  else:
   self.size = np.array([ max(nodes[:,0]) - min(nodes[:,0]),\
                  max(nodes[:,1]) - min(nodes[:,1]),0])

  self.nodes = nodes
  #import elements and create map
  self.side_elem_map = {}
  self.elem_side_map = {}
  f.readline()
  f.readline()
  n_tot = int(f.readline())
  self.sides = []
  self.elems = []
  b_sides = []
  nr = []

  self.node_side_map = {}
  self.side_node_map = {}
  self.node_elem_map = {}
  self.side_list = {}
 

  for n in range(n_tot):
   tmp = f.readline().split()

   #Get sides------------------------------------------------------------
   if self.dim == 3 and int(tmp[1]) == 2: #2D area
     n = sorted([int(tmp[5])-1,int(tmp[6])-1,int(tmp[7])-1])

   if self.dim == 2 and int(tmp[1]) == 1: #1D area
     n = sorted([int(tmp[5])-1,int(tmp[6])-1])

   b_sides.append(n)
   nr.append(int(tmp[3]))



   if self.dim == 3 and int(tmp[1]) == 4: #3D Elem
     node_indexes = [int(tmp[5])-1,int(tmp[6])-1,int(tmp[7])-1,int(tmp[8])-1]
     n = sorted(node_indexes)
     self.elems.append(n)
     perm_n =[[n[0],n[1],n[2]],\
             [n[0],n[1],n[3]],\
             [n[1],n[2],n[3]],\
             [n[0],n[2],n[3]]]
     self._update_map(perm_n,b_sides,nr,node_indexes)
     self.elem_region_map.update({len(self.elems)-1:self.blabels[int(tmp[3])]})
     self.region_elem_map.setdefault(self.blabels[int(tmp[3])],[]).append(len(self.elems)-1)


   if self.dim == 2 and int(tmp[1]) == 2: #2D Elem (triangle)
     node_indexes = [int(tmp[5])-1,int(tmp[6])-1,int(tmp[7])-1]
     n = sorted(node_indexes)
     self.elems.append(n)
     perm_n =[[n[0],n[1]],\
             [n[0],n[2]],\
             [n[1],n[2]]]

     self.elem_region_map.update({len(self.elems)-1:self.blabels[int(tmp[3])]})
     self.region_elem_map.setdefault(self.blabels[int(tmp[3])],[]).append(len(self.elems)-1)

     self._update_map(perm_n,b_sides,nr,node_indexes)

   if self.dim == 2 and int(tmp[1]) == 3: #2D Elem (quadrangle)
     n = [int(tmp[5])-1,int(tmp[6])-1,int(tmp[7])-1,int(tmp[8])-1]

     #n = sorted(node_indexes)
     self.elems.append(n)
     perm_n =[ sorted([n[0],n[1]]),\
              sorted([n[1],n[2]]),\
              sorted([n[2],n[3]]),\
              sorted([n[3],n[0]])]
     self._update_map(perm_n,b_sides,nr,n)

     self.elem_region_map.update({len(self.elems)-1:self.blabels[int(tmp[3])]})
     self.region_elem_map.setdefault(self.blabels[int(tmp[3])],[]).append(len(self.elems)-1)

  #------------------------------------------------------------
  #Set default for hot and cold
  self.side_list.setdefault('Hot',[])
  self.side_list.setdefault('Cold',[])




  #Apply Periodic Boundary Conditions
  self.side_list.update({'active':list(range(len(self.sides)))})
  self.side_periodicity = np.zeros((len(self.sides),2,3))
  group_1 = []
  group_2 = []
  self.exlude = []

  self.pairs = [] #global (all periodic pairs)

  self.side_list.setdefault('Boundary',[])
  self.side_list.setdefault('Interface',[])
  #self.node_list.setdefault('Interface',[])
  self.periodic_nodes = []




  for label in list(self.side_list.keys()):

   #Add Cold and Hot to Boundary
   #if label == "Hot" or label == "Cold":
   # tmp = self.side_list.setdefault('Boundary',[])+self.side_list[label]
   # self.side_list['Boundary'] = tmp

   if str(label.split('_')[0]) == 'Periodic':
    if not int(label.split('_')[1])%2==0:
     contact_1 = label
     contact_2 = 'Periodic_' + str(int(label.split('_')[1])+1)
     group_1 = self.side_list[contact_1]
     group_2 = self.side_list[contact_2]
     pairs = []

     #compute tangential unity vector
     tmp = self.nodes[self.sides[group_2[0]][0]] - self.nodes[self.sides[group_2[0]][1]]
     t = tmp/np.linalg.norm(tmp)
     n = len(group_1)
     for s1 in group_1:
      d_min = 1e6
      for s in group_2:
       c1 = self.compute_side_centroid(s1)
       c2 = self.compute_side_centroid(s)
       d = np.linalg.norm(c2-c1)
       if d < d_min:
        d_min = d
        pp = c1-c2
        s2 = s

      #if abs(np.dot(pp,self.nodes[self.sides[s1][0]]-self.nodes[self.sides[s1][1]]))>1e-4:
      pairs.append([s1,s2])
      self.side_periodicity[s1][1] = pp
      self.side_periodicity[s2][1] = -pp

      if np.linalg.norm(self.nodes[self.sides[s1][0]] - self.nodes[self.sides[s2][0]]) == np.linalg.norm(pp):
       self.periodic_nodes.append([self.sides[s1][0],self.sides[s2][0]])
       self.periodic_nodes.append([self.sides[s1][1],self.sides[s2][1]])
      else:
       self.periodic_nodes.append([self.sides[s1][0],self.sides[s2][1]])
       self.periodic_nodes.append([self.sides[s1][1],self.sides[s2][0]])


     plot_sides = False
     if plot_sides:
      for s in pairs:
       c1 = self.compute_side_centroid(s[0])
       c2 = self.compute_side_centroid(s[1])
       #plot([c1[0],c2[0]],[c1[1],c2[1]],color='r')
      #show()
     #Amend map
     for s in pairs:
      s1 = s[0]
      s2 = s[1]

      #Change side in elem 2
      elem2 = self.side_elem_map[s2][0]
      index = self.elem_side_map[elem2].index(s2)
      self.elem_side_map[elem2][index] = s1
      self.side_elem_map[s1].append(elem2)

      #To get hflux sides right-----
      self.side_elem_map[s2].append(self.side_elem_map[s1][0])

      #-------------------------

      #Delete s2 from active list
      self.side_list['active'].remove(s2)


      self.exlude.append(s2)
      tmp = self.side_list.setdefault('Inactive',[]) + [s2]
      self.side_list['Inactive'] = tmp

     #Polish sides
     tmp = self.side_list.setdefault('Periodic',[])+self.side_list[contact_1]
     self.side_list['Periodic'] = tmp


     del self.side_list[contact_1]
     del self.side_list[contact_2]
     self.pairs += pairs

  #Create boundary_elements--------------------

  self.boundary_elements = []
  self.interface_elements = []

  self.node_list = {}
  boundary_nodes = []
  for ll in self.side_list['Boundary']:
   self.boundary_elements.append(self.side_elem_map[ll][0])
   for node in self.sides[ll]:
    if not node in boundary_nodes:
     boundary_nodes.append(node)
  self.node_list.update({'Boundary':boundary_nodes})

  interface_nodes = []
  self.interface_elements = []
  for ll in self.side_list['Interface']:
   self.interface_elements.append(self.side_elem_map[ll][0])
   self.interface_elements.append(self.side_elem_map[ll][1])
   for node in self.sides[ll]:
    if not node in interface_nodes:
     interface_nodes.append(node)
  self.node_list.update({'Interface':interface_nodes})


  #print(self.node_list['Interface'])


  #delete MESH-----
  #a=subprocess.check_output(['rm','-f','mesh.geo'])
  #a=subprocess.check_output(['rm','-f','mesh.msh'])

  #----------------


 def get_elems_from_side(self,ll):

  return self.side_elem_map[ll]


 def _update_map(self,perm_n,b_sides,nr,node_indexes):


    #UPDATE NODE_ELEM_MAP-------------------
    for n in node_indexes:
     self.node_elem_map.setdefault(n,[]).append(len(self.elems)-1)
    #---------------------------------------

    for new_n in perm_n :
     #Use the node_side_map---------
     index = -1
     sides = []
     for k in new_n:
      if index == -1:
       if k in self.node_side_map.keys():
        for s in self.node_side_map[k]:
         if np.allclose(self.sides[s],new_n):
          index = s
          break

     if index == -1:
      self.sides.append(new_n)
      index = len(self.sides)-1

      #add boundary--------------------------------------
      try:
       k = b_sides.index(new_n)

       self.side_list.setdefault(self.blabels[nr[k]],[]).append(index)
      except:
       a = 1
      #------------------------------------------------------
      for i in range(len(new_n)):
       self.node_side_map.setdefault(new_n[i],[]).append(index)
       self.side_node_map.setdefault(index,[]).append(new_n[i])

     #------------------------------------
     self.side_elem_map.setdefault(index,[]).append(len(self.elems)-1)
     self.elem_side_map.setdefault(len(self.elems)-1,[]).append(index)





 def get_elem_boundary_area_normal(self,elem) :

    sides = self.elem_side_map[elem]
    is_here = 0
    for s in sides:
     if s in self.side_list['Boundary'] :
      normal = self.compute_side_normal(elem,s)
      is_here = 1
      break
    if is_here == 0 : print('ERROR: side is not a boundary')
    return normal


 def compute_side_normals(self):

  self.side_normals = np.zeros((len(self.sides),2,3))
  for s in range(len(self.sides)):
   elems = self.side_elem_map[s]
   self.side_normals[s][0] = self.compute_side_normal(elems[0],s)
   if len(elems)>1:
    self.side_normals[s][1] = self.compute_side_normal(elems[1],s)

 def compute_side_normal(self,ne,ns):

  #Get generic normal--------------------
  v1 = self.nodes[self.sides[ns][1]]-self.nodes[self.sides[ns][0]]
  if self.dim == 3:
   v2 = self.nodes[self.sides[ns][2]]-self.nodes[self.sides[ns][1]]
  else :
   v2 = np.array([0,0,1])
  v = np.cross(v1,v2)
  normal = v/np.linalg.norm(v)
  #-------------------------------------

  #Check sign
  ind = self.side_elem_map[ns].index(ne)
  c_el   = self.compute_centroid(self.elems[ne])
  c_side = self.compute_centroid(self.sides[ns]) -  self.side_periodicity[ns][ind]

  c = (c_side - c_el)
  if np.dot(normal,c) < 0: normal = - normal


  return normal

 def get_n_elems(self):
  return len(self.elems)

 def compute_side_area(self,ns):

  p = self.nodes[self.sides[ns]]
  if self.dim == 2:
    return np.linalg.norm(p[1]-p[0])
  else:
   v = np.cross(p[1]-p[0],p[1]-p[2])
   normal = v/np.linalg.norm(v)
   tmp = np.zeros(3)
   for i in range(len(p)):
    vi1 = p[i]
    vi2 = p[(i+1)%len(p)]
    tmp += np.cross(vi1, vi2)

   result = np.dot(tmp,normal)
   return abs(result/2)


 def write_vtk(self,file_name,data) :

   points = self.nodes
   el = self.elems
   cells = []
   for k in range(len(el)):
    if self.dim == 3:
     cells.append(el[k][0:4])
    else:
     cells.append(el[k][0:3])

   if self.dim == 3:
    vtk = VtkData(UnstructuredGrid(points,tetra=cells),data)
   else:
    vtk = VtkData(UnstructuredGrid(points,triangle=cells),data)
   vtk.tofile(file_name,'ascii')


 def get_side_normal(self,ne,ns):
  return self.side_normals[ns,ne]


 def get_elem_centroid(self,ne):
  return self.elem_centroids[ne]

 def get_side_centroid(self,ns):
  return self.side_centroids[ns]



 def compute_side_centroids(self):

  self.side_centroids = np.zeros((len(self.sides),3))
  for s in range(len(self.sides)):
   self.side_centroids[s] = self.compute_side_centroid(s)

 def compute_side_centroid(self,ns):

  nodes = self.nodes[self.sides[ns]]

  centroid = np.zeros(3)
  for p in nodes:
   centroid += p
  return centroid/len(nodes)



 def compute_elem_centroids(self):

  self.elem_centroids = np.zeros((len(self.elems),3))
  for elem in range(len(self.elems)):
   self.elem_centroids[elem] = self.compute_elem_centroid(elem)


 def compute_elem_centroid(self,ne):
  nodes = self.nodes[self.elems[ne]]
  centroid = np.zeros(3)
  for p in nodes:
   centroid += p

  return centroid/len(nodes)

 def compute_elem_volume(self,kc1):

  if self.dim == 3: #Assuming Tetraedron
   ns = self.elems[kc1]
   m = np.ones((4,4))
   m[0,0:3] = self.nodes[ns[0]]
   m[1,0:3] = self.nodes[ns[1]]
   m[2,0:3] = self.nodes[ns[2]]
   m[3,0:3] = self.nodes[ns[3]]
   return abs(1.0/6.0 * np.linalg.det(m))

  if self.dim == 2: #Assuming Tetraedron

   points = self.nodes[self.elems[kc1]]
   x = []; y= []
   for p in points:
    x.append(p[0])
    y.append(p[1])

   return 0.5*np.abs(np.dot(x,np.roll(y,1))-np.dot(y,np.roll(x,1)))


   return volume


 def get_side_periodic_value(self,elem,elem_p) :

    ll = self.get_side_between_two_elements(elem,elem_p)
    ind = self.side_elem_map[ll].index(elem)
    return self.side_periodic_value[ll][ind]


 def get_next_elem_centroid(self,elem,side):

  centroid1 = self.get_elem_centroid(elem)
  elem2 = self.get_neighbor_elem(elem,side)
  centroid2 = self.get_elem_centroid(elem2)
  ind1 = self.side_elem_map[side].index(elem)
  ind2 = self.side_elem_map[side].index(elem2)
  centroid = centroid2 - self.side_periodicity[side][ind1] + self.side_periodicity[side][ind2]
  return centroid

 def cross_interface(self,p1,p2):

  line1 = LineString([p1,p2])
  for ll in self.side_list['Interface']:
   p3 = self.nodes[self.sides[ll][0]][:2]
   p4 = self.nodes[self.sides[ll][1]][:2]
   line2 = LineString([p3,p4])
   tmp = line1.intersection(line2)
   if isinstance(tmp,shapely.geometry.Point):
     return np.array([tmp.x,tmp.y])
  return []


 def compute_centroid(self,side):

  node = np.zeros(3)
  for p in side:
   node += self.nodes[p]

  return node/len(side)

 def get_elem_extended_neighbors(self,elem):
  #This also includes the element get_distance_between_centroids_of_two_elements_from_side
  neighbors = []
  for n in self.elems[elem]:
   for elem2 in self.node_elem_map[n]:
    if not elem2 in neighbors: neighbors.append(elem2)
    #for m in self.elems[elem2]:
    # for elem3 in self.node_elem_map[m]:
    #    if not elem3 in neighbors: neighbors.append(elem3)

  return neighbors

 def get_elem_neighbors(self,elem):

  neighbors = []
  for ll in self.elem_side_map[elem]:
   neighbors.append(self.get_neighbor_elem(elem,ll))
  return neighbors

 def get_neighbor_elem(self,elem,ll) :

    if not (elem in self.side_elem_map[ll]) : print('error')

    for tmp in self.side_elem_map[ll] :
       if not (tmp == elem) :
         return tmp


 def compute_interpolation_weigths(self):

  self.interp_weigths = {}# np.zeros(len(self.sides))
  for ll in self.side_list['active']:
   if not (ll in (self.side_list['Boundary'] + self.side_list['Hot'] + self.side_list['Cold'])) :
    e0 = self.side_elem_map[ll][0]
    e1 = self.side_elem_map[ll][1]
    P0 = self.get_elem_centroid(e0)
    P1 = self.get_next_elem_centroid(e0,ll)
    #---------------------------------------------------------------
    if self.dim == 3:
     #from here: http://geomalgorithms.com/a05-_intersect-1.html
     u = P1 - P0
     n = self.get_side_normal(1,ll)
     node = self.nodes[self.sides[ll][0]]
     w = P0 - node
     s = -np.dot(n,w)/np.dot(n,u)
    else: #dim = 2
     P2 = self.nodes[self.sides[ll][0]]
     P3 = self.nodes[self.sides[ll][1]]
     den = (P3[1] - P2[1])*(P1[0]-P0[0])-(P3[0]-P2[0])*(P1[1]-P0[1])
     num = (P3[0] - P2[0])*(P0[1]-P2[1])-(P3[1]-P2[1])*(P0[0]-P2[0])
     a = num/den
     P = P0 + a * (P1-P0)
     if a < 0.0 or a > 1.0 :
      print(P0)
      print(P1)
      print('ERROR in the skew parameter')
      return
     dist = np.linalg.norm(P1-P0)
     d = np.linalg.norm(P - P1)
     s = d/dist
     #---------------------------------------------------------------
    self.interp_weigths.update({ll:s})
   else:
    #self.interp_weigths.update({ll:0.0})
    self.interp_weigths.update({ll:1.0})


 def get_region_from_elem(self,elem):
  return self.elem_region_map[elem]


 def get_interpolation_weigths(self,ll,elem):

  #return self.interp_weigths[ll]
  if self.side_elem_map[ll][0] == elem:
   return self.interp_weigths[ll]
  else:
   return 1.0-self.interp_weigths[ll]

 def compute_boundary_condition_data(self):

    gradir = self.direction

    if gradir == 0:
     flux_dir = [1,0,0]
     length = self.size[0]

    if gradir == 1:
     flux_dir = [0,1,0]
     length = self.size[1]

    if gradir == 2:
     flux_dir = [0,0,1]
     length = self.size[2]

    delta = 1e-2
    nsides = len(self.sides)
    flux_sides = [] #where flux goes
    side_value = np.zeros(nsides)

    tmp = self.side_list.setdefault('Periodic',[]) + self.exlude


    for kl,ll in enumerate(tmp) :
     normal = self.compute_side_normal(self.side_elem_map[ll][0],ll)
     tmp = np.dot(normal,flux_dir)
     if tmp > delta : #either negative or positive
      flux_sides.append(ll)

     if tmp < - delta :
        side_value[ll] = -1.0
     if tmp > delta :
        side_value[ll] = +1.0
        
    side_periodic_value = np.zeros((nsides,2))

   
    n_el = len(self.elems)
    B = sparse.DOK((n_el,n_el),dtype=float32)
    
    #ll = self.get_side_between_two_elements(elem,elem_p)
    #ind = self.side_elem_map[ll].index(elem)
    #return self.side_periodic_value[ll][ind]

     
    if len(self.side_list.setdefault('Periodic',[])) > 0:
     for side in self.pairs:
      side_periodic_value[side[0]][0] = side_value[side[0]]
      side_periodic_value[side[0]][1] = side_value[side[1]]
      

      elem1,elem2 = self.side_elem_map[side[0]]
      B[elem1,elem2] = side_value[side[0]]#*self.get_elem_volume(elem1)
      B[elem2,elem1] = side_value[side[1]]#*self.get_elem_volume(elem2)


    self.area_flux = abs(np.dot(flux_dir,self.c_areas))
    self.flux_sides = flux_sides
    self.side_periodic_value = side_periodic_value
    self.B = B.to_coo()



 def compute_least_square_weigths(self):

   n_el = len(self.elems)

   nd = len(self.elems[0])
   diff_dist = np.zeros((n_el,nd,self.dim))


   for ll in self.side_list['active'] :

    elems = self.side_elem_map[ll]
    kc1 = elems[0]
    c1 = self.compute_elem_centroid(kc1)
    ind1 = self.elem_side_map[kc1].index(ll)

    if not ll in (self.side_list['Boundary'] + self.side_list['Hot'] + self.side_list['Cold']):

     #Diff in the distance
     kc2 = elems[1]
     c2 = self.get_next_elem_centroid(kc1,ll)
     ind2 = self.elem_side_map[kc2].index(ll)
     dist = c2-c1

     for i in range(self.dim):
      diff_dist[kc1][ind1][i] = dist[i]
      diff_dist[kc2][ind2][i] = -dist[i]
    else :
     dist = self.compute_side_centroid(ll) - c1
     for i in range(self.dim):
      diff_dist[kc1][ind1][i] = dist[i]

   #Compute weights
   self.weigths = []
   for tmp in diff_dist :
    self.weigths.append(np.dot(np.linalg.inv(np.dot(np.transpose(tmp),tmp)),np.transpose(tmp)  ))


 def compute_grad(self,temp,lattice_temp =[]) :


   if len(lattice_temp) == 0: lattice_temp = temp

   n_el = len(self.elems)
   nd = len(self.elems[0])

   if self.dim == 2:
    diff_temp = np.zeros((n_el,nd))
   else:
    diff_temp = np.zeros((n_el,4))

   gradT = np.zeros((n_el,3))

   #Diff in the temp
   for ll in self.side_list['active'] :

    elems = self.side_elem_map[ll]
    kc1 = elems[0]
    #c1 = self.compute_elem_centroid(kc1)
    c1 = self.get_elem_centroid(kc1)
    ind1 = self.elem_side_map[kc1].index(ll)

    if not ll in (self.side_list['Boundary'] + self.side_list['Hot'] + self.side_list['Cold']) :

     kc2 = elems[1]
     ind2 = self.elem_side_map[kc2].index(ll)

     temp_1 = temp[kc1]
     temp_2 = temp[kc2] + self.get_side_periodic_value(kc2,kc1)
     diff_t = temp_2 - temp_1

     diff_temp[kc1][ind1]  = diff_t
     diff_temp[kc2][ind2]  = -diff_t
    else :
     if ll in self.side_list['Hot'] :
      diff_temp[kc1][ind1]  = 0.5-temp[kc1]

     if ll in self.side_list['Cold'] :
      diff_temp[kc1][ind1]  = -0.5-temp[kc1]

     if ll in self.side_list['Boundary'] :
      diff_temp[kc1][ind1]  = lattice_temp[kc1]-temp[kc1]

   for k in range(n_el) :
    tmp = np.dot(self.weigths[k],diff_temp[k])
    gradT[k][0] = tmp[0] #THESE HAS TO BE POSITIVE
    gradT[k][1] = tmp[1]
    if self.dim == 3:
     gradT[k][2] = tmp[2]

   return gradT




from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range
from past.builtins import basestring
from builtins import object
from past.utils import old_div
import urllib.request, urllib.parse, urllib.error


def main():

 urllib.request.urlretrieve('https://www.dropbox.com/s/adaeafjnzipitu0/example.py?dl=1','example.py')



from mpi4py import MPI
from pyvtk import *
import numpy as np
from matplotlib.tri import Triangulation

class WriteVtk(object):

 def __init__(self,argv):

  self.mesh = argv['Geometry']


  self.Nx = argv.setdefault('repeat_x',1)
  self.Ny = argv.setdefault('repeat_y',1)
  self.Nz = argv.setdefault('repeat_z',1)
  self.data = []
  self.label = []


 def add_variable(self,variable,**argv):

  self.data.append(variable)
  self.label.append(argv['label'])

 def cell_to_node(self,data):


   #n_nodes = len(self.mesh.nodes)

   #print(len(self.mesh.node_list['Interface']))
   delta = 2e-6*np.ones(3)
   #delta[2] = 0.0

   add_nodes = True
   if add_nodes:

    for node in self.mesh.node_list['Interface']:

     self.mesh.nodes = np.append(self.mesh.nodes,[self.mesh.nodes[node]],axis=0)
     node_elem_map_1 = []
     node_elem_map_2 = []
     for elem in self.mesh.node_elem_map[node]:
      if self.mesh.get_region_from_elem(elem) == 'Matrix':
        node_elem_map_1.append(elem)
      else:
        node_elem_map_2.append(elem)
        index = self.mesh.elems[elem].index(node)
        self.mesh.elems[elem][index]= len(self.mesh.nodes)-1


     self.mesh.node_elem_map[node] = node_elem_map_1
     self.mesh.node_elem_map.update({len(self.mesh.nodes)-1:node_elem_map_2})

   #for node  in self.mesh.node_list['Interface']:
#     print(len(self.mesh.node_elem_map[node]))

   #quit()
   n_nodes = len(self.mesh.nodes)
    #-----------------------
    #quit()
   n_col = len(np.shape(data))
   if n_col == 2: n_col = 3
   node_data = np.zeros((n_nodes,n_col))

   conn = np.zeros(n_nodes)
   for n in self.mesh.node_elem_map.keys():
    for elem in self.mesh.node_elem_map[n]:
     node_data[n] += data[elem]
     conn[n] +=1

   for n in range(n_nodes):
    node_data[n] /= conn[n]

   return node_data

 def repeat_nodes_and_data(self,data_uc,increment):

   periodic_nodes = np.array(self.mesh.periodic_nodes)
   periodic_1 = [i[0] for i in periodic_nodes]
   periodic_2 = [i[1] for i in periodic_nodes]

   #----------------------------------------------
   nodes_uc = self.mesh.nodes

   n_nodes = len(nodes_uc)

   nodes = []
   data = []
   #add nodes---
   corr = np.zeros(n_nodes*self.Nx*self.Ny*self.Nz,dtype=np.int)
   index = 0
   for nx in range(self.Nx):
    for ny in range(self.Ny):
     for nz in range(self.Nz):
      dd = nx*increment[0] + ny*increment[1] + nz*increment[2]
      index = nx * self.Ny * self.Nz * n_nodes +  ny *  self.Nz * n_nodes + nz * n_nodes
      P = [round(self.mesh.size[0]*nx,3),round(self.mesh.size[1]*ny,3),round(self.mesh.size[2]*nz,3)]
      for n in range(n_nodes):
       #index += 1
       node_trial = [round(nodes_uc[n,0],4) + P[0]-(self.Nx-1)*0.5*self.mesh.size[0],\
                     round(nodes_uc[n,1],4) + P[1]-(self.Nx-1)*0.5*self.mesh.size[1],\
                     round(nodes_uc[n,2],4) + P[2]-(self.Nx-1)*0.5*self.mesh.size[2]]

       #if not node_trial in nodes:
       nodes.append(node_trial)
       data.append(data_uc[n]-dd)
       corr[index+n] = len(nodes)-1

       #else:
    #    corr[index+n] = nodes.index(node_trial)


   cells_uc = self.mesh.elems

   cells = []
   for nx in range(self.Nx):
    for ny in range(self.Ny):
     for nz in range(self.Nz):
      P = [self.mesh.size[0]*nx,self.mesh.size[1]*ny,self.mesh.size[2]*nz]
      index = nx * self.Ny * self.Nz * n_nodes +  ny *  self.Nz * n_nodes + nz * n_nodes
      #add cells
      for gg in cells_uc:
       tmp = []
       for n in gg:
        tmp.append(corr[n+index])
       cells.append(tmp)



   return np.array(nodes),np.array(cells),np.array(data)

 def get_node_data(self,variable):

 
    node_data = self.cell_to_node(variable)
    is_scalar = len(list(np.shape(variable))) == 1
    increment = [0,0,0]
    tmp = [[1,0,0],[0,1,0],[0,0,1]]
    if is_scalar:
     increment = tmp[self.mesh.direction]
    nodes,cells,data = self.repeat_nodes_and_data(node_data,increment)

    #return Triangulation(np.array(self.mesh.nodes)[:,0],np.array(self.nodes)[:,1], triangles=self.mesh.elems, mask=None),node_data,self.mesh.nodes



    return Triangulation(nodes[:,0],nodes[:,1], triangles=cells, mask=None),data,nodes



 def write_vtk(self,filename='output.vtk'):

  if MPI.COMM_WORLD.Get_rank() == 0:
   stored_data = {}


   output = []
   strc = 'PointData('
   for n,variable in enumerate(self.data) :

    is_scalar = len(list(np.shape(variable))) == 1

    node_data = self.cell_to_node(variable)
    #stored_data.update({self.label[n]:{variable:node_data[:,0],'label':self.label[n]}})

    #Get increment for plotting-------------
    increment = [0,0,0]
    tmp = [[1,0,0],[0,1,0],[0,0,1]]
    if is_scalar:
     increment = tmp[self.mesh.direction]
    #---------------------------------------
    ss = '''self.repeat_nodes_and_data(node_data,increment)'''
    nodes,cells,tmp = eval(ss)
    output.append(tmp)
    if is_scalar:
      strc += r'''Scalars(output[''' + str(n) + r'''],name =' ''' + self.label[n] +  r''' ')'''
    else:
      strc += r'''Vectors(output[''' + str(n) + r'''],name =' ''' + self.label[n] +  r''' ')'''

    if n == len(self.data)-1:
     strc += ')'
    else:
     strc += ','
 
   data=eval(strc)

   if self.mesh.dim == 3:
    vtk = VtkData(UnstructuredGrid(nodes,tetra=cells),data)
   else :
    if len(cells[0]) == 3:
     vtk = VtkData(UnstructuredGrid(nodes,triangle=cells),data)
    else:
     vtk = VtkData(PolyData(points = nodes,polygons = cells),data)
   vtk.tofile(filename,'ascii')
